/* SPDX-License-Identifier: GPL-2.0 */
/* THIS FILE WAS AUTOGENERATED BY `jevents.py arch=none model=none` ! */

#include <pmu-events/pmu-events.h>
#include "util/header.h"
#include "util/pmu.h"
#include <string.h>
#include <stddef.h>

struct compact_pmu_event {
	int offset;
};

struct pmu_table_entry {
	const struct compact_pmu_event *entries;
	uint32_t num_entries;
	struct compact_pmu_event pmu_name;
};

/* clang-format off */
static const char *const big_c_string =
/* offset=0 */
"default_core\000"
/* offset=13 */
"l1-dcache\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=99 */
"l1-dcache-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=190 */
"l1-dcache-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=286 */
"l1-dcache-load-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=387 */
"l1-dcache-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=482 */
"l1-dcache-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=580 */
"l1-dcache-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00000\000\000\000\000\000"
/* offset=682 */
"l1-dcache-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=782 */
"l1-dcache-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00000\000\000\000\000\000"
/* offset=874 */
"l1-dcache-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=971 */
"l1-dcache-loads-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1073 */
"l1-dcache-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1169 */
"l1-dcache-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1268 */
"l1-dcache-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=1371 */
"l1-dcache-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=1472 */
"l1-dcache-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1563 */
"l1-dcache-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1659 */
"l1-dcache-read-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1760 */
"l1-dcache-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1855 */
"l1-dcache-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=1953 */
"l1-dcache-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=2055 */
"l1-dcache-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=2155 */
"l1-dcache-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=2252 */
"l1-dcache-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=2354 */
"l1-dcache-store-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=2461 */
"l1-dcache-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=2562 */
"l1-dcache-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=2666 */
"l1-dcache-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00000\000\000\000\000\000"
/* offset=2770 */
"l1-dcache-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=2872 */
"l1-dcache-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00000\000\000\000\000\000"
/* offset=2970 */
"l1-dcache-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3073 */
"l1-dcache-stores-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3181 */
"l1-dcache-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3283 */
"l1-dcache-stores-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3388 */
"l1-dcache-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=3493 */
"l1-dcache-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=3596 */
"l1-dcache-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3693 */
"l1-dcache-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3795 */
"l1-dcache-write-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=3902 */
"l1-dcache-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=4003 */
"l1-dcache-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=4107 */
"l1-dcache-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=4211 */
"l1-dcache-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=4313 */
"l1-dcache-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=4416 */
"l1-dcache-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=4524 */
"l1-dcache-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=4637 */
"l1-dcache-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=4744 */
"l1-dcache-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=4854 */
"l1-dcache-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00000\000\000\000\000\000"
/* offset=4964 */
"l1-dcache-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=5072 */
"l1-dcache-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00000\000\000\000\000\000"
/* offset=5177 */
"l1-dcache-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=5287 */
"l1-dcache-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=5402 */
"l1-dcache-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=5511 */
"l1-dcache-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=5623 */
"l1-dcache-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=5735 */
"l1-dcache-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=5845 */
"l1-dcache-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=5956 */
"l1-dcache-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6072 */
"l1-dcache-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6193 */
"l1-dcache-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6308 */
"l1-dcache-speculative-read-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6426 */
"l1-dcache-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=6544 */
"l1-dcache-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=6660 */
"l1-dcache-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6771 */
"l1-dcache-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=6887 */
"l1-dcache-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=7008 */
"l1-dcache-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=7123 */
"l1-dcache-speculative-load-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=7241 */
"l1-dcache-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=7359 */
"l1-dcache-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=7475 */
"l1-dcache-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=7566 */
"l1-dcache-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=7662 */
"l1-dcache-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=7752 */
"l1-dcache-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=7845 */
"l1-dcache-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=7942 */
"l1-dcache-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=8037 */
"l1-d\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8118 */
"l1-d-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8204 */
"l1-d-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8295 */
"l1-d-load-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8391 */
"l1-d-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8481 */
"l1-d-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8574 */
"l1-d-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=8671 */
"l1-d-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=8766 */
"l1-d-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8853 */
"l1-d-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=8945 */
"l1-d-loads-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9042 */
"l1-d-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9133 */
"l1-d-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9227 */
"l1-d-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=9325 */
"l1-d-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=9421 */
"l1-d-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9507 */
"l1-d-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9598 */
"l1-d-read-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9694 */
"l1-d-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9784 */
"l1-d-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=9877 */
"l1-d-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=9974 */
"l1-d-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=10069 */
"l1-d-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10161 */
"l1-d-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10258 */
"l1-d-store-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10360 */
"l1-d-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10456 */
"l1-d-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10555 */
"l1-d-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=10654 */
"l1-d-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=10751 */
"l1-d-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10844 */
"l1-d-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=10942 */
"l1-d-stores-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11045 */
"l1-d-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11142 */
"l1-d-stores-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11242 */
"l1-d-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=11342 */
"l1-d-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=11440 */
"l1-d-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11532 */
"l1-d-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11629 */
"l1-d-write-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11731 */
"l1-d-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11827 */
"l1-d-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=11926 */
"l1-d-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=12025 */
"l1-d-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=12122 */
"l1-d-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12220 */
"l1-d-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12323 */
"l1-d-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12431 */
"l1-d-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12533 */
"l1-d-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12638 */
"l1-d-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=12743 */
"l1-d-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=12846 */
"l1-d-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=12946 */
"l1-d-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13051 */
"l1-d-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13161 */
"l1-d-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13265 */
"l1-d-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13372 */
"l1-d-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=13479 */
"l1-d-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=13584 */
"l1-d-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13690 */
"l1-d-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13801 */
"l1-d-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=13917 */
"l1-d-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14027 */
"l1-d-speculative-read-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14140 */
"l1-d-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=14253 */
"l1-d-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=14364 */
"l1-d-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14470 */
"l1-d-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14581 */
"l1-d-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14697 */
"l1-d-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14807 */
"l1-d-speculative-load-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=14920 */
"l1-d-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=15033 */
"l1-d-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=15144 */
"l1-d-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15230 */
"l1-d-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15321 */
"l1-d-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15406 */
"l1-d-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15494 */
"l1-d-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=15586 */
"l1-d-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=15676 */
"l1d\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15756 */
"l1d-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15841 */
"l1d-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=15931 */
"l1d-load-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16026 */
"l1d-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16115 */
"l1d-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16207 */
"l1d-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=16303 */
"l1d-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=16397 */
"l1d-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16483 */
"l1d-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16574 */
"l1d-loads-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16670 */
"l1d-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16760 */
"l1d-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=16853 */
"l1d-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=16950 */
"l1d-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=17045 */
"l1d-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=17130 */
"l1d-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=17220 */
"l1d-read-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=17315 */
"l1d-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=17404 */
"l1d-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=17496 */
"l1d-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=17592 */
"l1d-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=17686 */
"l1d-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=17777 */
"l1d-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=17873 */
"l1d-store-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=17974 */
"l1d-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18069 */
"l1d-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18167 */
"l1d-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=18265 */
"l1d-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=18361 */
"l1d-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18453 */
"l1d-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18550 */
"l1d-stores-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18652 */
"l1d-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18748 */
"l1d-stores-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=18847 */
"l1d-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=18946 */
"l1d-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=19043 */
"l1d-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=19134 */
"l1d-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=19230 */
"l1d-write-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=19331 */
"l1d-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=19426 */
"l1d-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=19524 */
"l1d-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=19622 */
"l1d-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=19718 */
"l1d-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=19815 */
"l1d-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=19917 */
"l1d-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20024 */
"l1d-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20125 */
"l1d-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20229 */
"l1d-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=20333 */
"l1d-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=20435 */
"l1d-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20534 */
"l1d-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20638 */
"l1d-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20747 */
"l1d-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20850 */
"l1d-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=20956 */
"l1d-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=21062 */
"l1d-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=21166 */
"l1d-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=21271 */
"l1d-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=21381 */
"l1d-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=21496 */
"l1d-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=21605 */
"l1d-speculative-read-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=21717 */
"l1d-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=21829 */
"l1d-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=21939 */
"l1d-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=22044 */
"l1d-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=22154 */
"l1d-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=22269 */
"l1d-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=22378 */
"l1d-speculative-load-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=22490 */
"l1d-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=22602 */
"l1d-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=22712 */
"l1d-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=22797 */
"l1d-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=22887 */
"l1d-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=22971 */
"l1d-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23058 */
"l1d-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=23149 */
"l1d-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=23238 */
"l1-data\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23322 */
"l1-data-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23411 */
"l1-data-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23505 */
"l1-data-load-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23604 */
"l1-data-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23697 */
"l1-data-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=23793 */
"l1-data-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=23893 */
"l1-data-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=23991 */
"l1-data-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24081 */
"l1-data-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24176 */
"l1-data-loads-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24276 */
"l1-data-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24370 */
"l1-data-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24467 */
"l1-data-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=24568 */
"l1-data-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=24667 */
"l1-data-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24756 */
"l1-data-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24850 */
"l1-data-read-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=24949 */
"l1-data-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=25042 */
"l1-data-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=25138 */
"l1-data-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=25238 */
"l1-data-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=25336 */
"l1-data-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=25431 */
"l1-data-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=25531 */
"l1-data-store-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=25636 */
"l1-data-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=25735 */
"l1-data-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=25837 */
"l1-data-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=25939 */
"l1-data-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=26039 */
"l1-data-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26135 */
"l1-data-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26236 */
"l1-data-stores-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26342 */
"l1-data-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26442 */
"l1-data-stores-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26545 */
"l1-data-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=26648 */
"l1-data-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=26749 */
"l1-data-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26844 */
"l1-data-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=26944 */
"l1-data-write-reference\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=27049 */
"l1-data-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=27148 */
"l1-data-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-config=0x100\000\00010\000\000\000\000\000"
/* offset=27250 */
"l1-data-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=27352 */
"l1-data-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cache-config=0x10100\000\00010\000\000\000\000\000"
/* offset=27452 */
"l1-data-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=27553 */
"l1-data-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=27659 */
"l1-data-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=27770 */
"l1-data-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=27875 */
"l1-data-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=27983 */
"l1-data-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=28091 */
"l1-data-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=28197 */
"l1-data-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=28300 */
"l1-data-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=28408 */
"l1-data-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=28521 */
"l1-data-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=28628 */
"l1-data-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=28738 */
"l1-data-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=28848 */
"l1-data-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=28956 */
"l1-data-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29065 */
"l1-data-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29179 */
"l1-data-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29298 */
"l1-data-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29411 */
"l1-data-speculative-read-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29527 */
"l1-data-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=29643 */
"l1-data-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=29757 */
"l1-data-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29866 */
"l1-data-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=29980 */
"l1-data-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=30099 */
"l1-data-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=30212 */
"l1-data-speculative-load-access\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-cache-config=0x200\000\00010\000\000\000\000\000"
/* offset=30328 */
"l1-data-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=30444 */
"l1-data-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses\000legacy-cache-config=0x10200\000\00010\000\000\000\000\000"
/* offset=30558 */
"l1-data-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=30647 */
"l1-data-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=30741 */
"l1-data-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=30829 */
"l1-data-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0\000\00010\000\000\000\000\000"
/* offset=30920 */
"l1-data-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=31015 */
"l1-data-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-config=0x10000\000\00010\000\000\000\000\000"
/* offset=31108 */
"l1-icache\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31201 */
"l1-icache-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31299 */
"l1-icache-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31402 */
"l1-icache-load-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31510 */
"l1-icache-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31612 */
"l1-icache-load-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=31717 */
"l1-icache-load-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00000\000\000\000\000\000"
/* offset=31826 */
"l1-icache-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=31933 */
"l1-icache-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00000\000\000\000\000\000"
/* offset=32032 */
"l1-icache-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32136 */
"l1-icache-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32245 */
"l1-icache-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32348 */
"l1-icache-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32454 */
"l1-icache-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=32564 */
"l1-icache-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=32672 */
"l1-icache-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32770 */
"l1-icache-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32873 */
"l1-icache-read-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=32981 */
"l1-icache-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=33083 */
"l1-icache-read-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=33188 */
"l1-icache-read-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=33297 */
"l1-icache-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=33404 */
"l1-icache-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=33514 */
"l1-icache-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=33629 */
"l1-icache-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=33749 */
"l1-icache-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=33863 */
"l1-icache-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=33980 */
"l1-icache-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00000\000\000\000\000\000"
/* offset=34097 */
"l1-icache-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=34212 */
"l1-icache-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00000\000\000\000\000\000"
/* offset=34324 */
"l1-icache-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=34441 */
"l1-icache-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=34563 */
"l1-icache-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=34679 */
"l1-icache-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=34798 */
"l1-icache-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=34917 */
"l1-icache-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=35034 */
"l1-icache-speculative-read\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=35152 */
"l1-icache-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=35275 */
"l1-icache-speculative-read-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=35403 */
"l1-icache-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=35525 */
"l1-icache-speculative-read-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=35650 */
"l1-icache-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=35775 */
"l1-icache-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=35898 */
"l1-icache-speculative-load\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=36016 */
"l1-icache-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=36139 */
"l1-icache-speculative-load-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=36267 */
"l1-icache-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=36389 */
"l1-icache-speculative-load-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=36514 */
"l1-icache-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=36639 */
"l1-icache-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=36762 */
"l1-icache-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=36860 */
"l1-icache-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=36963 */
"l1-icache-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37060 */
"l1-icache-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37160 */
"l1-icache-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=37264 */
"l1-icache-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=37366 */
"l1-i\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37454 */
"l1-i-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37547 */
"l1-i-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37645 */
"l1-i-load-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37748 */
"l1-i-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37845 */
"l1-i-load-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=37945 */
"l1-i-load-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=38049 */
"l1-i-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=38151 */
"l1-i-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38245 */
"l1-i-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38344 */
"l1-i-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38448 */
"l1-i-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38546 */
"l1-i-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38647 */
"l1-i-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=38752 */
"l1-i-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=38855 */
"l1-i-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=38948 */
"l1-i-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=39046 */
"l1-i-read-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=39149 */
"l1-i-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=39246 */
"l1-i-read-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=39346 */
"l1-i-read-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=39450 */
"l1-i-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=39552 */
"l1-i-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=39657 */
"l1-i-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=39767 */
"l1-i-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=39882 */
"l1-i-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=39991 */
"l1-i-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40103 */
"l1-i-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=40215 */
"l1-i-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=40325 */
"l1-i-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40432 */
"l1-i-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40544 */
"l1-i-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40661 */
"l1-i-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40772 */
"l1-i-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=40886 */
"l1-i-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=41000 */
"l1-i-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=41112 */
"l1-i-speculative-read\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=41225 */
"l1-i-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=41343 */
"l1-i-speculative-read-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=41466 */
"l1-i-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=41583 */
"l1-i-speculative-read-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=41703 */
"l1-i-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=41823 */
"l1-i-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=41941 */
"l1-i-speculative-load\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=42054 */
"l1-i-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=42172 */
"l1-i-speculative-load-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=42295 */
"l1-i-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=42412 */
"l1-i-speculative-load-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=42532 */
"l1-i-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=42652 */
"l1-i-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=42770 */
"l1-i-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=42863 */
"l1-i-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=42961 */
"l1-i-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43053 */
"l1-i-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43148 */
"l1-i-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=43247 */
"l1-i-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=43344 */
"l1i\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43431 */
"l1i-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43523 */
"l1i-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43620 */
"l1i-load-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43722 */
"l1i-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43818 */
"l1i-load-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=43917 */
"l1i-load-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=44020 */
"l1i-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=44121 */
"l1i-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44214 */
"l1i-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44312 */
"l1i-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44415 */
"l1i-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44512 */
"l1i-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44612 */
"l1i-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=44716 */
"l1i-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=44818 */
"l1i-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=44910 */
"l1i-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=45007 */
"l1i-read-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=45109 */
"l1i-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=45205 */
"l1i-read-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=45304 */
"l1i-read-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=45407 */
"l1i-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=45508 */
"l1i-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=45612 */
"l1i-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=45721 */
"l1i-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=45835 */
"l1i-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=45943 */
"l1i-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46054 */
"l1i-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=46165 */
"l1i-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=46274 */
"l1i-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46380 */
"l1i-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46491 */
"l1i-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46607 */
"l1i-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46717 */
"l1i-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=46830 */
"l1i-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=46943 */
"l1i-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=47054 */
"l1i-speculative-read\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47166 */
"l1i-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47283 */
"l1i-speculative-read-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47405 */
"l1i-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47521 */
"l1i-speculative-read-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47640 */
"l1i-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=47759 */
"l1i-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=47876 */
"l1i-speculative-load\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=47988 */
"l1i-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=48105 */
"l1i-speculative-load-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=48227 */
"l1i-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=48343 */
"l1i-speculative-load-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=48462 */
"l1i-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=48581 */
"l1i-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=48698 */
"l1i-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=48790 */
"l1i-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=48887 */
"l1i-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=48978 */
"l1i-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49072 */
"l1i-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=49170 */
"l1i-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=49266 */
"l1-instruction\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49364 */
"l1-instruction-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49467 */
"l1-instruction-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49575 */
"l1-instruction-load-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49688 */
"l1-instruction-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49795 */
"l1-instruction-load-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=49905 */
"l1-instruction-load-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=50019 */
"l1-instruction-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=50131 */
"l1-instruction-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=50235 */
"l1-instruction-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=50344 */
"l1-instruction-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=50458 */
"l1-instruction-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=50566 */
"l1-instruction-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=50677 */
"l1-instruction-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=50792 */
"l1-instruction-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=50905 */
"l1-instruction-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=51008 */
"l1-instruction-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=51116 */
"l1-instruction-read-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=51229 */
"l1-instruction-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=51336 */
"l1-instruction-read-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=51446 */
"l1-instruction-read-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=51560 */
"l1-instruction-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=51672 */
"l1-instruction-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=51787 */
"l1-instruction-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=51907 */
"l1-instruction-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52032 */
"l1-instruction-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52151 */
"l1-instruction-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52273 */
"l1-instruction-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=52395 */
"l1-instruction-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=52515 */
"l1-instruction-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52632 */
"l1-instruction-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52754 */
"l1-instruction-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=52881 */
"l1-instruction-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53002 */
"l1-instruction-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53126 */
"l1-instruction-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=53250 */
"l1-instruction-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=53372 */
"l1-instruction-speculative-read\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53495 */
"l1-instruction-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53623 */
"l1-instruction-speculative-read-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53756 */
"l1-instruction-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=53883 */
"l1-instruction-speculative-read-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54013 */
"l1-instruction-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=54143 */
"l1-instruction-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=54271 */
"l1-instruction-speculative-load\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54394 */
"l1-instruction-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54522 */
"l1-instruction-speculative-load-reference\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54655 */
"l1-instruction-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54782 */
"l1-instruction-speculative-load-access\000legacy cache\000Level 1 instruction cache prefetch accesses\000legacy-cache-config=0x201\000\00010\000\000\000\000\000"
/* offset=54912 */
"l1-instruction-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=55042 */
"l1-instruction-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch misses\000legacy-cache-config=0x10201\000\00010\000\000\000\000\000"
/* offset=55170 */
"l1-instruction-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=55273 */
"l1-instruction-reference\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=55381 */
"l1-instruction-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=55483 */
"l1-instruction-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-config=1\000\00010\000\000\000\000\000"
/* offset=55588 */
"l1-instruction-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=55697 */
"l1-instruction-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cache-config=0x10001\000\00010\000\000\000\000\000"
/* offset=55804 */
"llc\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=55882 */
"llc-load\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=55965 */
"llc-load-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56053 */
"llc-load-reference\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56146 */
"llc-load-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56233 */
"llc-load-access\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56323 */
"llc-load-misses\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00000\000\000\000\000\000"
/* offset=56417 */
"llc-load-miss\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=56509 */
"llc-loads\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00000\000\000\000\000\000"
/* offset=56593 */
"llc-loads-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56682 */
"llc-loads-reference\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56776 */
"llc-loads-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56864 */
"llc-loads-access\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=56955 */
"llc-loads-misses\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=57050 */
"llc-loads-miss\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=57143 */
"llc-read\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=57226 */
"llc-read-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=57314 */
"llc-read-reference\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=57407 */
"llc-read-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=57494 */
"llc-read-access\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=57584 */
"llc-read-misses\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=57678 */
"llc-read-miss\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=57770 */
"llc-store\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=57859 */
"llc-store-refs\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=57953 */
"llc-store-reference\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58052 */
"llc-store-ops\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58145 */
"llc-store-access\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58241 */
"llc-store-misses\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00000\000\000\000\000\000"
/* offset=58337 */
"llc-store-miss\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=58431 */
"llc-stores\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00000\000\000\000\000\000"
/* offset=58521 */
"llc-stores-refs\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58616 */
"llc-stores-reference\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58716 */
"llc-stores-ops\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58810 */
"llc-stores-access\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=58907 */
"llc-stores-misses\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=59004 */
"llc-stores-miss\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=59099 */
"llc-write\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=59188 */
"llc-write-refs\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=59282 */
"llc-write-reference\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=59381 */
"llc-write-ops\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=59474 */
"llc-write-access\000legacy cache\000Last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=59570 */
"llc-write-misses\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=59666 */
"llc-write-miss\000legacy cache\000Last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=59760 */
"llc-prefetch\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=59855 */
"llc-prefetch-refs\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=59955 */
"llc-prefetch-reference\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60060 */
"llc-prefetch-ops\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60159 */
"llc-prefetch-access\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60261 */
"llc-prefetch-misses\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00000\000\000\000\000\000"
/* offset=60363 */
"llc-prefetch-miss\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=60463 */
"llc-prefetches\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00000\000\000\000\000\000"
/* offset=60560 */
"llc-prefetches-refs\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60662 */
"llc-prefetches-reference\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60769 */
"llc-prefetches-ops\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60870 */
"llc-prefetches-access\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=60974 */
"llc-prefetches-misses\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=61078 */
"llc-prefetches-miss\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=61180 */
"llc-speculative-read\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=61283 */
"llc-speculative-read-refs\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=61391 */
"llc-speculative-read-reference\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=61504 */
"llc-speculative-read-ops\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=61611 */
"llc-speculative-read-access\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=61721 */
"llc-speculative-read-misses\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=61831 */
"llc-speculative-read-miss\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=61939 */
"llc-speculative-load\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=62042 */
"llc-speculative-load-refs\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=62150 */
"llc-speculative-load-reference\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=62263 */
"llc-speculative-load-ops\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=62370 */
"llc-speculative-load-access\000legacy cache\000Last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=62480 */
"llc-speculative-load-misses\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=62590 */
"llc-speculative-load-miss\000legacy cache\000Last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=62698 */
"llc-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=62781 */
"llc-reference\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=62869 */
"llc-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=62951 */
"llc-access\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63036 */
"llc-misses\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=63125 */
"llc-miss\000legacy cache\000Last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=63212 */
"l2\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63309 */
"l2-load\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63411 */
"l2-load-refs\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63518 */
"l2-load-reference\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63630 */
"l2-load-ops\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63736 */
"l2-load-access\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=63845 */
"l2-load-misses\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=63958 */
"l2-load-miss\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=64069 */
"l2-loads\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64172 */
"l2-loads-refs\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64280 */
"l2-loads-reference\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64393 */
"l2-loads-ops\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64500 */
"l2-loads-access\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64610 */
"l2-loads-misses\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=64724 */
"l2-loads-miss\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=64836 */
"l2-read\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=64938 */
"l2-read-refs\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=65045 */
"l2-read-reference\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=65157 */
"l2-read-ops\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=65263 */
"l2-read-access\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=65372 */
"l2-read-misses\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=65485 */
"l2-read-miss\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=65596 */
"l2-store\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=65704 */
"l2-store-refs\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=65817 */
"l2-store-reference\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=65935 */
"l2-store-ops\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66047 */
"l2-store-access\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66162 */
"l2-store-misses\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=66277 */
"l2-store-miss\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=66390 */
"l2-stores\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66499 */
"l2-stores-refs\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66613 */
"l2-stores-reference\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66732 */
"l2-stores-ops\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66845 */
"l2-stores-access\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=66961 */
"l2-stores-misses\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=67077 */
"l2-stores-miss\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=67191 */
"l2-write\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=67299 */
"l2-write-refs\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=67412 */
"l2-write-reference\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=67530 */
"l2-write-ops\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=67642 */
"l2-write-access\000legacy cache\000Level 2 (or higher) last level cache write accesses\000legacy-cache-config=0x102\000\00010\000\000\000\000\000"
/* offset=67757 */
"l2-write-misses\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=67872 */
"l2-write-miss\000legacy cache\000Level 2 (or higher) last level cache write misses\000legacy-cache-config=0x10102\000\00010\000\000\000\000\000"
/* offset=67985 */
"l2-prefetch\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68099 */
"l2-prefetch-refs\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68218 */
"l2-prefetch-reference\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68342 */
"l2-prefetch-ops\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68460 */
"l2-prefetch-access\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68581 */
"l2-prefetch-misses\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=68702 */
"l2-prefetch-miss\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=68821 */
"l2-prefetches\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=68937 */
"l2-prefetches-refs\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69058 */
"l2-prefetches-reference\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69184 */
"l2-prefetches-ops\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69304 */
"l2-prefetches-access\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69427 */
"l2-prefetches-misses\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=69550 */
"l2-prefetches-miss\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=69671 */
"l2-speculative-read\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69793 */
"l2-speculative-read-refs\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=69920 */
"l2-speculative-read-reference\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70052 */
"l2-speculative-read-ops\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70178 */
"l2-speculative-read-access\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70307 */
"l2-speculative-read-misses\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=70436 */
"l2-speculative-read-miss\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=70563 */
"l2-speculative-load\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70685 */
"l2-speculative-load-refs\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70812 */
"l2-speculative-load-reference\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=70944 */
"l2-speculative-load-ops\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=71070 */
"l2-speculative-load-access\000legacy cache\000Level 2 (or higher) last level cache prefetch accesses\000legacy-cache-config=0x202\000\00010\000\000\000\000\000"
/* offset=71199 */
"l2-speculative-load-misses\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=71328 */
"l2-speculative-load-miss\000legacy cache\000Level 2 (or higher) last level cache prefetch misses\000legacy-cache-config=0x10202\000\00010\000\000\000\000\000"
/* offset=71455 */
"l2-refs\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=71557 */
"l2-reference\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=71664 */
"l2-ops\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=71765 */
"l2-access\000legacy cache\000Level 2 (or higher) last level cache read accesses\000legacy-cache-config=2\000\00010\000\000\000\000\000"
/* offset=71869 */
"l2-misses\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=71977 */
"l2-miss\000legacy cache\000Level 2 (or higher) last level cache read misses\000legacy-cache-config=0x10002\000\00010\000\000\000\000\000"
/* offset=72083 */
"dtlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72154 */
"dtlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72230 */
"dtlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72311 */
"dtlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72397 */
"dtlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72477 */
"dtlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72560 */
"dtlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00000\000\000\000\000\000"
/* offset=72647 */
"dtlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=72732 */
"dtlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00000\000\000\000\000\000"
/* offset=72809 */
"dtlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72891 */
"dtlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=72978 */
"dtlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73059 */
"dtlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73143 */
"dtlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=73231 */
"dtlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=73317 */
"dtlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73393 */
"dtlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73474 */
"dtlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73560 */
"dtlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73640 */
"dtlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=73723 */
"dtlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=73810 */
"dtlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=73895 */
"dtlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=73977 */
"dtlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74064 */
"dtlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74156 */
"dtlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74242 */
"dtlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74331 */
"dtlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00000\000\000\000\000\000"
/* offset=74420 */
"dtlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=74507 */
"dtlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00000\000\000\000\000\000"
/* offset=74590 */
"dtlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74678 */
"dtlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74771 */
"dtlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74858 */
"dtlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=74948 */
"dtlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=75038 */
"dtlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=75126 */
"dtlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=75208 */
"dtlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=75295 */
"dtlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=75387 */
"dtlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=75473 */
"dtlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=75562 */
"dtlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=75651 */
"dtlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=75738 */
"dtlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=75826 */
"dtlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=75919 */
"dtlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76017 */
"dtlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76109 */
"dtlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76204 */
"dtlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00000\000\000\000\000\000"
/* offset=76299 */
"dtlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=76392 */
"dtlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00000\000\000\000\000\000"
/* offset=76482 */
"dtlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76577 */
"dtlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76677 */
"dtlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76771 */
"dtlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=76868 */
"dtlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=76965 */
"dtlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=77060 */
"dtlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77156 */
"dtlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77257 */
"dtlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77363 */
"dtlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77463 */
"dtlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77566 */
"dtlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=77669 */
"dtlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=77770 */
"dtlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77866 */
"dtlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=77967 */
"dtlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=78073 */
"dtlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=78173 */
"dtlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=78276 */
"dtlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=78379 */
"dtlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=78480 */
"dtlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=78556 */
"dtlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=78637 */
"dtlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=78712 */
"dtlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=78790 */
"dtlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=78872 */
"dtlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=78952 */
"d-tlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79024 */
"d-tlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79101 */
"d-tlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79183 */
"d-tlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79270 */
"d-tlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79351 */
"d-tlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79435 */
"d-tlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=79523 */
"d-tlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=79609 */
"d-tlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79687 */
"d-tlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79770 */
"d-tlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79858 */
"d-tlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=79940 */
"d-tlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80025 */
"d-tlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=80114 */
"d-tlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=80201 */
"d-tlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80278 */
"d-tlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80360 */
"d-tlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80447 */
"d-tlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80528 */
"d-tlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=80612 */
"d-tlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=80700 */
"d-tlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=80786 */
"d-tlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=80869 */
"d-tlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=80957 */
"d-tlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81050 */
"d-tlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81137 */
"d-tlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81227 */
"d-tlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=81317 */
"d-tlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=81405 */
"d-tlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81489 */
"d-tlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81578 */
"d-tlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81672 */
"d-tlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81760 */
"d-tlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=81851 */
"d-tlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=81942 */
"d-tlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=82031 */
"d-tlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=82114 */
"d-tlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=82202 */
"d-tlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=82295 */
"d-tlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=82382 */
"d-tlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=82472 */
"d-tlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=82562 */
"d-tlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=82650 */
"d-tlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=82739 */
"d-tlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=82833 */
"d-tlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=82932 */
"d-tlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83025 */
"d-tlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83121 */
"d-tlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=83217 */
"d-tlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=83311 */
"d-tlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83402 */
"d-tlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83498 */
"d-tlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83599 */
"d-tlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83694 */
"d-tlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=83792 */
"d-tlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=83890 */
"d-tlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=83986 */
"d-tlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84083 */
"d-tlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84185 */
"d-tlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84292 */
"d-tlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84393 */
"d-tlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84497 */
"d-tlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=84601 */
"d-tlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=84703 */
"d-tlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84800 */
"d-tlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=84902 */
"d-tlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=85009 */
"d-tlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=85110 */
"d-tlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=85214 */
"d-tlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=85318 */
"d-tlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=85420 */
"d-tlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=85497 */
"d-tlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=85579 */
"d-tlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=85655 */
"d-tlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=85734 */
"d-tlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=85817 */
"d-tlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=85898 */
"data-tlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=85973 */
"data-tlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86053 */
"data-tlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86138 */
"data-tlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86228 */
"data-tlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86312 */
"data-tlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86399 */
"data-tlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=86490 */
"data-tlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=86579 */
"data-tlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86660 */
"data-tlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86746 */
"data-tlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86837 */
"data-tlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=86922 */
"data-tlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87010 */
"data-tlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=87102 */
"data-tlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=87192 */
"data-tlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87272 */
"data-tlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87357 */
"data-tlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87447 */
"data-tlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87531 */
"data-tlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=87618 */
"data-tlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=87709 */
"data-tlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=87798 */
"data-tlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=87884 */
"data-tlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=87975 */
"data-tlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88071 */
"data-tlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88161 */
"data-tlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88254 */
"data-tlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=88347 */
"data-tlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=88438 */
"data-tlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88525 */
"data-tlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88617 */
"data-tlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88714 */
"data-tlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88805 */
"data-tlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=88899 */
"data-tlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=88993 */
"data-tlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=89085 */
"data-tlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=89171 */
"data-tlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=89262 */
"data-tlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=89358 */
"data-tlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=89448 */
"data-tlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x103\000\00010\000\000\000\000\000"
/* offset=89541 */
"data-tlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=89634 */
"data-tlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=0x10103\000\00010\000\000\000\000\000"
/* offset=89725 */
"data-tlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=89817 */
"data-tlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=89914 */
"data-tlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90016 */
"data-tlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90112 */
"data-tlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90211 */
"data-tlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=90310 */
"data-tlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=90407 */
"data-tlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90501 */
"data-tlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90600 */
"data-tlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90704 */
"data-tlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90802 */
"data-tlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=90903 */
"data-tlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=91004 */
"data-tlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=91103 */
"data-tlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91203 */
"data-tlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91308 */
"data-tlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91418 */
"data-tlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91522 */
"data-tlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91629 */
"data-tlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=91736 */
"data-tlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=91841 */
"data-tlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=91941 */
"data-tlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=92046 */
"data-tlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=92156 */
"data-tlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=92260 */
"data-tlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-config=0x203\000\00010\000\000\000\000\000"
/* offset=92367 */
"data-tlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=92474 */
"data-tlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-config=0x10203\000\00010\000\000\000\000\000"
/* offset=92579 */
"data-tlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=92659 */
"data-tlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=92744 */
"data-tlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=92823 */
"data-tlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\00010\000\000\000\000\000"
/* offset=92905 */
"data-tlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=92991 */
"data-tlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003\000\00010\000\000\000\000\000"
/* offset=93075 */
"itlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93153 */
"itlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93236 */
"itlb-load-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93324 */
"itlb-load-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93417 */
"itlb-load-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93504 */
"itlb-load-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93594 */
"itlb-load-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00000\000\000\000\000\000"
/* offset=93688 */
"itlb-load-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=93780 */
"itlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00000\000\000\000\000\000"
/* offset=93864 */
"itlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=93953 */
"itlb-loads-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94047 */
"itlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94135 */
"itlb-loads-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94226 */
"itlb-loads-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=94321 */
"itlb-loads-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=94414 */
"itlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94497 */
"itlb-read-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94585 */
"itlb-read-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94678 */
"itlb-read-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94765 */
"itlb-read-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=94855 */
"itlb-read-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=94949 */
"itlb-read-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=95041 */
"itlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95124 */
"itlb-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95212 */
"itlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95294 */
"itlb-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95379 */
"itlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=95468 */
"itlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=95555 */
"i-tlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95634 */
"i-tlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95718 */
"i-tlb-load-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95807 */
"i-tlb-load-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95901 */
"i-tlb-load-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=95989 */
"i-tlb-load-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96080 */
"i-tlb-load-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=96175 */
"i-tlb-load-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=96268 */
"i-tlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96353 */
"i-tlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96443 */
"i-tlb-loads-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96538 */
"i-tlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96627 */
"i-tlb-loads-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96719 */
"i-tlb-loads-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=96815 */
"i-tlb-loads-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=96909 */
"i-tlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=96993 */
"i-tlb-read-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97082 */
"i-tlb-read-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97176 */
"i-tlb-read-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97264 */
"i-tlb-read-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97355 */
"i-tlb-read-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=97450 */
"i-tlb-read-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=97543 */
"i-tlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97627 */
"i-tlb-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97716 */
"i-tlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97799 */
"i-tlb-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=97885 */
"i-tlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=97975 */
"i-tlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=98063 */
"instruction-tlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98152 */
"instruction-tlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98246 */
"instruction-tlb-load-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98345 */
"instruction-tlb-load-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98449 */
"instruction-tlb-load-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98547 */
"instruction-tlb-load-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98648 */
"instruction-tlb-load-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=98753 */
"instruction-tlb-load-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=98856 */
"instruction-tlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=98951 */
"instruction-tlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99051 */
"instruction-tlb-loads-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99156 */
"instruction-tlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99255 */
"instruction-tlb-loads-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99357 */
"instruction-tlb-loads-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=99463 */
"instruction-tlb-loads-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=99567 */
"instruction-tlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99661 */
"instruction-tlb-read-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99760 */
"instruction-tlb-read-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99864 */
"instruction-tlb-read-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=99962 */
"instruction-tlb-read-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=100063 */
"instruction-tlb-read-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=100168 */
"instruction-tlb-read-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=100271 */
"instruction-tlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=100365 */
"instruction-tlb-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=100464 */
"instruction-tlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=100557 */
"instruction-tlb-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\000\00010\000\000\000\000\000"
/* offset=100653 */
"instruction-tlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=100753 */
"instruction-tlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=0x10004\000\00010\000\000\000\000\000"
/* offset=100851 */
"branch\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=100938 */
"branch-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101030 */
"branch-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101127 */
"branch-load-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101229 */
"branch-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101325 */
"branch-load-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101424 */
"branch-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00000\000\000\000\000\000"
/* offset=101527 */
"branch-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=101628 */
"branch-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00000\000\000\000\000\000"
/* offset=101721 */
"branch-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101819 */
"branch-loads-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=101922 */
"branch-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102019 */
"branch-loads-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102119 */
"branch-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=102223 */
"branch-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=102325 */
"branch-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102417 */
"branch-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102514 */
"branch-read-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102616 */
"branch-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102712 */
"branch-read-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=102811 */
"branch-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=102914 */
"branch-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=103015 */
"branch-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103107 */
"branch-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103204 */
"branch-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103295 */
"branch-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103389 */
"branch-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=103485 */
"branches-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103579 */
"branches-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103678 */
"branches-load-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103782 */
"branches-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103880 */
"branches-load-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=103981 */
"branches-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=104086 */
"branches-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=104189 */
"branches-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104284 */
"branches-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104384 */
"branches-loads-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104489 */
"branches-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104588 */
"branches-loads-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104690 */
"branches-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=104796 */
"branches-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=104900 */
"branches-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=104994 */
"branches-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105093 */
"branches-read-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105197 */
"branches-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105295 */
"branches-read-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105396 */
"branches-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=105501 */
"branches-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=105604 */
"branches-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105698 */
"branches-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105797 */
"branches-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105890 */
"branches-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=105986 */
"branches-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=106086 */
"branches-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=106184 */
"bpu\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106268 */
"bpu-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106357 */
"bpu-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106451 */
"bpu-load-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106550 */
"bpu-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106643 */
"bpu-load-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=106739 */
"bpu-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=106839 */
"bpu-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=106937 */
"bpu-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107027 */
"bpu-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107122 */
"bpu-loads-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107222 */
"bpu-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107316 */
"bpu-loads-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107413 */
"bpu-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=107514 */
"bpu-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=107613 */
"bpu-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107702 */
"bpu-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107796 */
"bpu-read-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107895 */
"bpu-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=107988 */
"bpu-read-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108084 */
"bpu-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=108184 */
"bpu-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=108282 */
"bpu-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108371 */
"bpu-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108465 */
"bpu-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108553 */
"bpu-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108644 */
"bpu-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=108739 */
"bpu-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=108832 */
"btb\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=108916 */
"btb-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109005 */
"btb-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109099 */
"btb-load-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109198 */
"btb-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109291 */
"btb-load-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109387 */
"btb-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=109487 */
"btb-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=109585 */
"btb-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109675 */
"btb-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109770 */
"btb-loads-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109870 */
"btb-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=109964 */
"btb-loads-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110061 */
"btb-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=110162 */
"btb-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=110261 */
"btb-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110350 */
"btb-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110444 */
"btb-read-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110543 */
"btb-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110636 */
"btb-read-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=110732 */
"btb-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=110832 */
"btb-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=110930 */
"btb-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111019 */
"btb-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111113 */
"btb-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111201 */
"btb-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111292 */
"btb-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=111387 */
"btb-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=111480 */
"bpc\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111564 */
"bpc-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111653 */
"bpc-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111747 */
"bpc-load-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111846 */
"bpc-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=111939 */
"bpc-load-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112035 */
"bpc-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=112135 */
"bpc-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=112233 */
"bpc-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112323 */
"bpc-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112418 */
"bpc-loads-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112518 */
"bpc-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112612 */
"bpc-loads-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112709 */
"bpc-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=112810 */
"bpc-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=112909 */
"bpc-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=112998 */
"bpc-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113092 */
"bpc-read-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113191 */
"bpc-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113284 */
"bpc-read-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113380 */
"bpc-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=113480 */
"bpc-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=113578 */
"bpc-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113667 */
"bpc-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113761 */
"bpc-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113849 */
"bpc-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-config=5\000\00010\000\000\000\000\000"
/* offset=113940 */
"bpc-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=114035 */
"bpc-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-config=0x10005\000\00010\000\000\000\000\000"
/* offset=114128 */
"node\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114203 */
"node-load\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114283 */
"node-load-refs\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114368 */
"node-load-reference\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114458 */
"node-load-ops\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114542 */
"node-load-access\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114629 */
"node-load-misses\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00000\000\000\000\000\000"
/* offset=114720 */
"node-load-miss\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=114809 */
"node-loads\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00000\000\000\000\000\000"
/* offset=114890 */
"node-loads-refs\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=114976 */
"node-loads-reference\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115067 */
"node-loads-ops\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115152 */
"node-loads-access\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115240 */
"node-loads-misses\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=115332 */
"node-loads-miss\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=115422 */
"node-read\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115502 */
"node-read-refs\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115587 */
"node-read-reference\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115677 */
"node-read-ops\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115761 */
"node-read-access\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=115848 */
"node-read-misses\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=115939 */
"node-read-miss\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=116028 */
"node-store\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116114 */
"node-store-refs\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116205 */
"node-store-reference\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116301 */
"node-store-ops\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116391 */
"node-store-access\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116484 */
"node-store-misses\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00000\000\000\000\000\000"
/* offset=116577 */
"node-store-miss\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00010\000\000\000\000\000"
/* offset=116668 */
"node-stores\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00000\000\000\000\000\000"
/* offset=116755 */
"node-stores-refs\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116847 */
"node-stores-reference\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=116944 */
"node-stores-ops\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117035 */
"node-stores-access\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117129 */
"node-stores-misses\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00010\000\000\000\000\000"
/* offset=117223 */
"node-stores-miss\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00010\000\000\000\000\000"
/* offset=117315 */
"node-write\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117401 */
"node-write-refs\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117492 */
"node-write-reference\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117588 */
"node-write-ops\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117678 */
"node-write-access\000legacy cache\000Local memory write accesses\000legacy-cache-config=0x106\000\00010\000\000\000\000\000"
/* offset=117771 */
"node-write-misses\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00010\000\000\000\000\000"
/* offset=117864 */
"node-write-miss\000legacy cache\000Local memory write misses\000legacy-cache-config=0x10106\000\00010\000\000\000\000\000"
/* offset=117955 */
"node-prefetch\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118047 */
"node-prefetch-refs\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118144 */
"node-prefetch-reference\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118246 */
"node-prefetch-ops\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118342 */
"node-prefetch-access\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118441 */
"node-prefetch-misses\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00000\000\000\000\000\000"
/* offset=118540 */
"node-prefetch-miss\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=118637 */
"node-prefetches\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00000\000\000\000\000\000"
/* offset=118731 */
"node-prefetches-refs\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118830 */
"node-prefetches-reference\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=118934 */
"node-prefetches-ops\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119032 */
"node-prefetches-access\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119133 */
"node-prefetches-misses\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=119234 */
"node-prefetches-miss\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=119333 */
"node-speculative-read\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119433 */
"node-speculative-read-refs\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119538 */
"node-speculative-read-reference\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119648 */
"node-speculative-read-ops\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119752 */
"node-speculative-read-access\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=119859 */
"node-speculative-read-misses\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=119966 */
"node-speculative-read-miss\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=120071 */
"node-speculative-load\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=120171 */
"node-speculative-load-refs\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=120276 */
"node-speculative-load-reference\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=120386 */
"node-speculative-load-ops\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=120490 */
"node-speculative-load-access\000legacy cache\000Local memory prefetch accesses\000legacy-cache-config=0x206\000\00010\000\000\000\000\000"
/* offset=120597 */
"node-speculative-load-misses\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=120704 */
"node-speculative-load-miss\000legacy cache\000Local memory prefetch misses\000legacy-cache-config=0x10206\000\00010\000\000\000\000\000"
/* offset=120809 */
"node-refs\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=120889 */
"node-reference\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=120974 */
"node-ops\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=121053 */
"node-access\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\00010\000\000\000\000\000"
/* offset=121135 */
"node-misses\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=121221 */
"node-miss\000legacy cache\000Local memory read misses\000legacy-cache-config=0x10006\000\00010\000\000\000\000\000"
/* offset=121305 */
"cpu-cycles\000legacy hardware\000Total cycles. Be wary of what happens during CPU frequency scaling [This event is an alias of cycles]\000legacy-hardware-config=0\000\00000\000\000\000\000\000"
/* offset=121467 */
"cycles\000legacy hardware\000Total cycles. Be wary of what happens during CPU frequency scaling [This event is an alias of cpu-cycles]\000legacy-hardware-config=0\000\00000\000\000\000\000\000"
/* offset=121629 */
"instructions\000legacy hardware\000Retired instructions. Be careful, these can be affected by various issues, most notably hardware interrupt counts\000legacy-hardware-config=1\000\00000\000\000\000\000\000"
/* offset=121805 */
"cache-references\000legacy hardware\000Cache accesses. Usually this indicates Last Level Cache accesses but this may vary depending on your CPU.  This may include prefetches and coherency messages; again this depends on the design of your CPU\000legacy-hardware-config=2\000\00000\000\000\000\000\000"
/* offset=122075 */
"cache-misses\000legacy hardware\000Cache misses. Usually this indicates Last Level Cache misses; this is intended to be used in conjunction with the PERF_COUNT_HW_CACHE_REFERENCES event to calculate cache miss rates\000legacy-hardware-config=3\000\00000\000\000\000\000\000"
/* offset=122318 */
"branches\000legacy hardware\000Retired branch instructions [This event is an alias of branch-instructions]\000legacy-hardware-config=4\000\00000\000\000\000\000\000"
/* offset=122452 */
"branch-instructions\000legacy hardware\000Retired branch instructions [This event is an alias of branches]\000legacy-hardware-config=4\000\00000\000\000\000\000\000"
/* offset=122586 */
"branch-misses\000legacy hardware\000Mispredicted branch instructions\000legacy-hardware-config=5\000\00000\000\000\000\000\000"
/* offset=122682 */
"bus-cycles\000legacy hardware\000Bus cycles, which can be different from total cycles\000legacy-hardware-config=6\000\00000\000\000\000\000\000"
/* offset=122795 */
"stalled-cycles-frontend\000legacy hardware\000Stalled cycles during issue [This event is an alias of idle-cycles-frontend]\000legacy-hardware-config=7\000\00000\000\000\000\000\000"
/* offset=122945 */
"idle-cycles-frontend\000legacy hardware\000Stalled cycles during issue [This event is an alias of stalled-cycles-fronted]\000legacy-hardware-config=7\000\00000\000\000\000\000\000"
/* offset=123094 */
"stalled-cycles-backend\000legacy hardware\000Stalled cycles during retirement [This event is an alias of idle-cycles-backend]\000legacy-hardware-config=8\000\00000\000\000\000\000\000"
/* offset=123247 */
"idle-cycles-backend\000legacy hardware\000Stalled cycles during retirement [This event is an alias of stalled-cycles-backend]\000legacy-hardware-config=8\000\00000\000\000\000\000\000"
/* offset=123400 */
"ref-cycles\000legacy hardware\000Total cycles; not affected by CPU frequency scaling\000legacy-hardware-config=9\000\00000\000\000\000\000\000"
/* offset=123512 */
"software\000"
/* offset=123521 */
"cpu-clock\000software\000Per-CPU high-resolution timer based event\000config=0\000\000001e-6msec\000\000\000\000\000"
/* offset=123607 */
"task-clock\000software\000Per-task high-resolution timer based event\000config=1\000\000001e-6msec\000\000\000\000\000"
/* offset=123695 */
"faults\000software\000Number of page faults [This event is an alias of page-faults]\000config=2\000\00000\000\000\000\000\000"
/* offset=123790 */
"page-faults\000software\000Number of page faults [This event is an alias of faults]\000config=2\000\00000\000\000\000\000\000"
/* offset=123885 */
"context-switches\000software\000Number of context switches [This event is an alias of cs]\000config=3\000\00000\000\000\000\000\000"
/* offset=123986 */
"cs\000software\000Number of context switches [This event is an alias of context-switches]\000config=3\000\00000\000\000\000\000\000"
/* offset=124087 */
"cpu-migrations\000software\000Number of times a process has migrated to a new CPU [This event is an alias of migrations]\000config=4\000\00000\000\000\000\000\000"
/* offset=124219 */
"migrations\000software\000Number of times a process has migrated to a new CPU [This event is an alias of cpu-migrations]\000config=4\000\00000\000\000\000\000\000"
/* offset=124351 */
"minor-faults\000software\000Number of minor page faults. Minor faults don't require I/O to handle\000config=5\000\00000\000\000\000\000\000"
/* offset=124460 */
"major-faults\000software\000Number of major page faults. Major faults require I/O to handle\000config=6\000\00000\000\000\000\000\000"
/* offset=124563 */
"alignment-faults\000software\000Number of kernel handled memory alignment faults\000config=7\000\00000\000\000\000\000\000"
/* offset=124655 */
"emulation-faults\000software\000Number of kernel handled unimplemented instruction faults handled through emulation\000config=8\000\00000\000\000\000\000\000"
/* offset=124782 */
"dummy\000software\000A placeholder event that doesn't count anything\000config=9\000\00000\000\000\000\000\000"
/* offset=124862 */
"bpf-output\000software\000An event used by BPF programs to write to the perf ring buffer\000config=0xa\000\00000\000\000\000\000\000"
/* offset=124964 */
"cgroup-switches\000software\000Number of context switches to a task in a different cgroup\000config=0xb\000\00000\000\000\000\000\000"
/* offset=125067 */
"tool\000"
/* offset=125072 */
"duration_time\000tool\000Wall clock interval time in nanoseconds\000config=1\000\00000\000\000\000\000\000"
/* offset=125148 */
"user_time\000tool\000User (non-kernel) time in nanoseconds\000config=2\000\00000\000\000\000\000\000"
/* offset=125218 */
"system_time\000tool\000System/kernel time in nanoseconds\000config=3\000\00000\000\000\000\000\000"
/* offset=125286 */
"has_pmem\000tool\0001 if persistent memory installed otherwise 0\000config=4\000\00000\000\000\000\000\000"
/* offset=125362 */
"num_cores\000tool\000Number of cores. A core consists of 1 or more thread, with each thread being associated with a logical Linux CPU\000config=5\000\00000\000\000\000\000\000"
/* offset=125507 */
"num_cpus\000tool\000Number of logical Linux CPUs. There may be multiple such CPUs on a core\000config=6\000\00000\000\000\000\000\000"
/* offset=125610 */
"num_cpus_online\000tool\000Number of online logical Linux CPUs. There may be multiple such CPUs on a core\000config=7\000\00000\000\000\000\000\000"
/* offset=125727 */
"num_dies\000tool\000Number of dies. Each die has 1 or more cores\000config=8\000\00000\000\000\000\000\000"
/* offset=125803 */
"num_packages\000tool\000Number of packages. Each package has 1 or more die\000config=9\000\00000\000\000\000\000\000"
/* offset=125889 */
"slots\000tool\000Number of functional units that in parallel can execute parts of an instruction\000config=0xa\000\00000\000\000\000\000\000"
/* offset=125999 */
"smt_on\000tool\0001 if simultaneous multithreading (aka hyperthreading) is enable otherwise 0\000config=0xb\000\00000\000\000\000\000\000"
/* offset=126106 */
"system_tsc_freq\000tool\000The amount a Time Stamp Counter (TSC) increases per second\000config=0xc\000\00000\000\000\000\000\000"
/* offset=126205 */
"core_wide\000tool\0001 if not SMT, if SMT are events being gathered on all SMT threads 1 otherwise 0\000config=0xd\000\00000\000\000\000\000\000"
/* offset=126319 */
"target_cpu\000tool\0001 if CPUs being analyzed, 0 if threads/processes\000config=0xe\000\00000\000\000\000\000\000"
/* offset=126403 */
"bp_l1_btb_correct\000branch\000L1 BTB Correction\000event=0x8a\000\00000\000\000\000\000\000"
/* offset=126465 */
"bp_l2_btb_correct\000branch\000L2 BTB Correction\000event=0x8b\000\00000\000\000\000\000\000"
/* offset=126527 */
"l3_cache_rd\000cache\000L3 cache access, read\000event=0x40\000\00000\000\000\000\000Attributable Level 3 cache access, read\000"
/* offset=126625 */
"segment_reg_loads.any\000other\000Number of segment register loads\000event=6,period=200000,umask=0x80\000\00000\000\000\000\000\000"
/* offset=126727 */
"dispatch_blocked.any\000other\000Memory cluster signals to block micro-op dispatch for any reason\000event=9,period=200000,umask=0x20\000\00000\000\000\000\000\000"
/* offset=126860 */
"eist_trans\000other\000Number of Enhanced Intel SpeedStep(R) Technology (EIST) transitions\000event=0x3a,period=200000\000\00000\000\000\000\000\000"
/* offset=126978 */
"hisi_sccl,ddrc\000"
/* offset=126993 */
"uncore_hisi_ddrc.flux_wcmd\000uncore\000DDRC write commands\000event=2\000\00000\000\000\000\000\000"
/* offset=127063 */
"uncore_cbox\000"
/* offset=127075 */
"unc_cbo_xsnp_response.miss_eviction\000uncore\000A cross-core snoop resulted from L3 Eviction which misses in some processor core\000event=0x22,umask=0x81\000\00000\000\000\000\000\000"
/* offset=127229 */
"event-hyphen\000uncore\000UNC_CBO_HYPHEN\000event=0xe0\000\00000\000\000\000\000\000"
/* offset=127283 */
"event-two-hyph\000uncore\000UNC_CBO_TWO_HYPH\000event=0xc0\000\00000\000\000\000\000\000"
/* offset=127341 */
"hisi_sccl,l3c\000"
/* offset=127355 */
"uncore_hisi_l3c.rd_hit_cpipe\000uncore\000Total read hits\000event=7\000\00000\000\000\000\000\000"
/* offset=127423 */
"uncore_imc_free_running\000"
/* offset=127447 */
"uncore_imc_free_running.cache_miss\000uncore\000Total cache misses\000event=0x12\000\00000\000\000\000\000\000"
/* offset=127527 */
"uncore_imc\000"
/* offset=127538 */
"uncore_imc.cache_hits\000uncore\000Total cache hits\000event=0x34\000\00000\000\000\000\000\000"
/* offset=127603 */
"uncore_sys_ddr_pmu\000"
/* offset=127622 */
"sys_ddr_pmu.write_cycles\000uncore\000ddr write-cycles event\000event=0x2b\000v8\00000\000\000\000\000\000"
/* offset=127698 */
"uncore_sys_ccn_pmu\000"
/* offset=127717 */
"sys_ccn_pmu.read_cycles\000uncore\000ccn read-cycles event\000config=0x2c\0000x01\00000\000\000\000\000\000"
/* offset=127794 */
"uncore_sys_cmn_pmu\000"
/* offset=127813 */
"sys_cmn_pmu.hnf_cache_miss\000uncore\000Counts total cache misses in first lookup result (high priority)\000eventid=1,type=5\000(434|436|43c|43a).*\00000\000\000\000\000\000"
/* offset=127956 */
"CPUs_utilized\000Default\000(software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@) / (duration_time * 1e9)\000\000Average CPU utilization\000\0001CPUs\000\000\000\000011"
/* offset=128142 */
"cs_per_second\000Default\000software@context\\-switches\\,name\\=context\\-switches@ * 1e9 / (software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@)\000\000Context switches per CPU second\000\0001cs/sec\000\000\000\000011"
/* offset=128375 */
"migrations_per_second\000Default\000software@cpu\\-migrations\\,name\\=cpu\\-migrations@ * 1e9 / (software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@)\000\000Process migrations to a new CPU per CPU second\000\0001migrations/sec\000\000\000\000011"
/* offset=128635 */
"page_faults_per_second\000Default\000software@page\\-faults\\,name\\=page\\-faults@ * 1e9 / (software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@)\000\000Page faults per CPU second\000\0001faults/sec\000\000\000\000011"
/* offset=128866 */
"insn_per_cycle\000Default\000instructions / cpu\\-cycles\000insn_per_cycle < 1\000Instructions Per Cycle\000\0001instructions\000\000\000\000001"
/* offset=128979 */
"stalled_cycles_per_instruction\000Default\000(max(stalled\\-cycles\\-frontend, stalled\\-cycles\\-backend) / instructions if has_event(stalled\\-cycles\\-frontend) & has_event(stalled\\-cycles\\-backend) else (stalled\\-cycles\\-frontend / instructions if has_event(stalled\\-cycles\\-frontend) else (stalled\\-cycles\\-backend / instructions if has_event(stalled\\-cycles\\-backend) else 0)))\000\000Max front or backend stalls per instruction\000\000\000\000\000\000001"
/* offset=129404 */
"frontend_cycles_idle\000Default\000(stalled\\-cycles\\-frontend / cpu\\-cycles if has_event(stalled\\-cycles\\-frontend) else 0)\000frontend_cycles_idle > 0.1\000Frontend stalls per cycle\000\000\000\000\000\000001"
/* offset=129583 */
"backend_cycles_idle\000Default\000(stalled\\-cycles\\-backend / cpu\\-cycles if has_event(stalled\\-cycles\\-backend) else 0)\000backend_cycles_idle > 0.2\000Backend stalls per cycle\000\000\000\000\000\000001"
/* offset=129757 */
"cycles_frequency\000Default\000cpu\\-cycles / (software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@)\000\000Cycles per CPU second\000\0001GHz\000\000\000\000011"
/* offset=129933 */
"branch_frequency\000Default\000branches / (software@cpu\\-clock\\,name\\=cpu\\-clock@ if #target_cpu else software@task\\-clock\\,name\\=task\\-clock@)\000\000Branches per CPU second\000\0001000M/sec\000\000\000\000011"
/* offset=130113 */
"branch_miss_rate\000Default\000branch\\-misses / branches\000branch_miss_rate > 0.05\000Branch miss rate\000\000100%\000\000\000\000001"
/* offset=130217 */
"l1d_miss_rate\000Default2\000L1\\-dcache\\-load\\-misses / L1\\-dcache\\-loads\000l1d_miss_rate > 0.05\000L1D  miss rate\000\000100%\000\000\000\000001"
/* offset=130333 */
"llc_miss_rate\000Default2\000LLC\\-load\\-misses / LLC\\-loads\000llc_miss_rate > 0.05\000LLC miss rate\000\000100%\000\000\000\000001"
/* offset=130434 */
"l1i_miss_rate\000Default3\000L1\\-icache\\-load\\-misses / L1\\-icache\\-loads\000l1i_miss_rate > 0.05\000L1I miss rate\000\000100%\000\000\000\000001"
/* offset=130549 */
"dtlb_miss_rate\000Default3\000dTLB\\-load\\-misses / dTLB\\-loads\000dtlb_miss_rate > 0.05\000dTLB miss rate\000\000100%\000\000\000\000001"
/* offset=130655 */
"itlb_miss_rate\000Default3\000iTLB\\-load\\-misses / iTLB\\-loads\000itlb_miss_rate > 0.05\000iTLB miss rate\000\000100%\000\000\000\000001"
/* offset=130761 */
"l1_prefetch_miss_rate\000Default4\000L1\\-dcache\\-prefetch\\-misses / L1\\-dcache\\-prefetches\000l1_prefetch_miss_rate > 0.05\000L1 prefetch miss rate\000\000100%\000\000\000\000001"
/* offset=130909 */
"CPI\000\0001 / IPC\000\000\000\000\000\000\000\000000"
/* offset=130932 */
"IPC\000group1\000inst_retired.any / cpu_clk_unhalted.thread\000\000\000\000\000\000\000\000000"
/* offset=130996 */
"Frontend_Bound_SMT\000\000idq_uops_not_delivered.core / (4 * (cpu_clk_unhalted.thread / 2 * (1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk)))\000\000\000\000\000\000\000\000000"
/* offset=131163 */
"dcache_miss_cpi\000\000l1d\\-loads\\-misses / inst_retired.any\000\000\000\000\000\000\000\000000"
/* offset=131228 */
"icache_miss_cycles\000\000l1i\\-loads\\-misses / inst_retired.any\000\000\000\000\000\000\000\000000"
/* offset=131296 */
"cache_miss_cycles\000group1\000dcache_miss_cpi + icache_miss_cycles\000\000\000\000\000\000\000\000000"
/* offset=131368 */
"DCache_L2_All_Hits\000\000l2_rqsts.demand_data_rd_hit + l2_rqsts.pf_hit + l2_rqsts.rfo_hit\000\000\000\000\000\000\000\000000"
/* offset=131463 */
"DCache_L2_All_Miss\000\000max(l2_rqsts.all_demand_data_rd - l2_rqsts.demand_data_rd_hit, 0) + l2_rqsts.pf_miss + l2_rqsts.rfo_miss\000\000\000\000\000\000\000\000000"
/* offset=131598 */
"DCache_L2_All\000\000DCache_L2_All_Hits + DCache_L2_All_Miss\000\000\000\000\000\000\000\000000"
/* offset=131663 */
"DCache_L2_Hits\000\000d_ratio(DCache_L2_All_Hits, DCache_L2_All)\000\000\000\000\000\000\000\000000"
/* offset=131732 */
"DCache_L2_Misses\000\000d_ratio(DCache_L2_All_Miss, DCache_L2_All)\000\000\000\000\000\000\000\000000"
/* offset=131803 */
"M1\000\000ipc + M2\000\000\000\000\000\000\000\000000"
/* offset=131826 */
"M2\000\000ipc + M1\000\000\000\000\000\000\000\000000"
/* offset=131849 */
"M3\000\0001 / M3\000\000\000\000\000\000\000\000000"
/* offset=131870 */
"L1D_Cache_Fill_BW\000\00064 * l1d.replacement / 1e9 / duration_time\000\000\000\000\000\000\000\000000"
;

static const struct compact_pmu_event pmu_events__common_default_core[] = {
	/* bpc\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-conf... */
	{ 111480 },
	/* bpc-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cac... */
	{ 113849 },
	/* bpc-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 111564 },
	/* bpc-load-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 111939 },
	/* bpc-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 112135 },
	/* bpc-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 112035 },
	/* bpc-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 111846 },
	/* bpc-load-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 111747 },
	/* bpc-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 111653 },
	/* bpc-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cach... */
	{ 112233 },
	/* bpc-loads-access\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 112612 },
	/* bpc-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-c... */
	{ 112810 },
	/* bpc-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy... */
	{ 112709 },
	/* bpc-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 112518 },
	/* bpc-loads-reference\000legacy cache\000Branch prediction unit read accesses\000l... */
	{ 112418 },
	/* bpc-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy... */
	{ 112323 },
	/* bpc-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-c... */
	{ 114035 },
	/* bpc-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache... */
	{ 113940 },
	/* bpc-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-... */
	{ 113761 },
	/* bpc-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 112909 },
	/* bpc-read-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 113284 },
	/* bpc-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 113480 },
	/* bpc-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 113380 },
	/* bpc-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 113191 },
	/* bpc-read-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 113092 },
	/* bpc-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 112998 },
	/* bpc-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 113667 },
	/* bpc-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 113578 },
	/* bpu\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-conf... */
	{ 106184 },
	/* bpu-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cac... */
	{ 108553 },
	/* bpu-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 106268 },
	/* bpu-load-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 106643 },
	/* bpu-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 106839 },
	/* bpu-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 106739 },
	/* bpu-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 106550 },
	/* bpu-load-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 106451 },
	/* bpu-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 106357 },
	/* bpu-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cach... */
	{ 106937 },
	/* bpu-loads-access\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 107316 },
	/* bpu-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-c... */
	{ 107514 },
	/* bpu-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy... */
	{ 107413 },
	/* bpu-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 107222 },
	/* bpu-loads-reference\000legacy cache\000Branch prediction unit read accesses\000l... */
	{ 107122 },
	/* bpu-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy... */
	{ 107027 },
	/* bpu-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-c... */
	{ 108739 },
	/* bpu-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache... */
	{ 108644 },
	/* bpu-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-... */
	{ 108465 },
	/* bpu-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 107613 },
	/* bpu-read-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 107988 },
	/* bpu-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 108184 },
	/* bpu-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 108084 },
	/* bpu-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 107895 },
	/* bpu-read-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 107796 },
	/* bpu-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 107702 },
	/* bpu-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 108371 },
	/* bpu-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 108282 },
	/* branch\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-c... */
	{ 100851 },
	/* branch-access\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 103295 },
	/* branch-instructions\000legacy hardware\000Retired branch instructions [This even... */
	{ 122452 },
	/* branch-load\000legacy cache\000Branch prediction unit read accesses\000legacy-ca... */
	{ 100938 },
	/* branch-load-access\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 101325 },
	/* branch-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy... */
	{ 101527 },
	/* branch-load-misses\000legacy cache\000Branch prediction unit read misses\000lega... */
	{ 101424 },
	/* branch-load-ops\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 101229 },
	/* branch-load-reference\000legacy cache\000Branch prediction unit read accesses\00... */
	{ 101127 },
	/* branch-load-refs\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 101030 },
	/* branch-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 101628 },
	/* branch-loads-access\000legacy cache\000Branch prediction unit read accesses\000l... */
	{ 102019 },
	/* branch-loads-miss\000legacy cache\000Branch prediction unit read misses\000legac... */
	{ 102223 },
	/* branch-loads-misses\000legacy cache\000Branch prediction unit read misses\000leg... */
	{ 102119 },
	/* branch-loads-ops\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 101922 },
	/* branch-loads-reference\000legacy cache\000Branch prediction unit read accesses\0... */
	{ 101819 },
	/* branch-loads-refs\000legacy cache\000Branch prediction unit read accesses\000leg... */
	{ 101721 },
	/* branch-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cach... */
	{ 103389 },
	/* branch-misses\000legacy hardware\000Mispredicted branch instructions\000legacy-h... */
	{ 122586 },
	/* branch-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cac... */
	{ 103204 },
	/* branch-read\000legacy cache\000Branch prediction unit read accesses\000legacy-ca... */
	{ 102325 },
	/* branch-read-access\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 102712 },
	/* branch-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy... */
	{ 102914 },
	/* branch-read-misses\000legacy cache\000Branch prediction unit read misses\000lega... */
	{ 102811 },
	/* branch-read-ops\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 102616 },
	/* branch-read-reference\000legacy cache\000Branch prediction unit read accesses\00... */
	{ 102514 },
	/* branch-read-refs\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 102417 },
	/* branch-reference\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 103107 },
	/* branch-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-ca... */
	{ 103015 },
	/* branches\000legacy hardware\000Retired branch instructions [This event is an ali... */
	{ 122318 },
	/* branches-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 105890 },
	/* branches-load\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 103485 },
	/* branches-load-access\000legacy cache\000Branch prediction unit read accesses\000... */
	{ 103880 },
	/* branches-load-miss\000legacy cache\000Branch prediction unit read misses\000lega... */
	{ 104086 },
	/* branches-load-misses\000legacy cache\000Branch prediction unit read misses\000le... */
	{ 103981 },
	/* branches-load-ops\000legacy cache\000Branch prediction unit read accesses\000leg... */
	{ 103782 },
	/* branches-load-reference\000legacy cache\000Branch prediction unit read accesses\... */
	{ 103678 },
	/* branches-load-refs\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 103579 },
	/* branches-loads\000legacy cache\000Branch prediction unit read accesses\000legacy... */
	{ 104189 },
	/* branches-loads-access\000legacy cache\000Branch prediction unit read accesses\00... */
	{ 104588 },
	/* branches-loads-miss\000legacy cache\000Branch prediction unit read misses\000leg... */
	{ 104796 },
	/* branches-loads-misses\000legacy cache\000Branch prediction unit read misses\000l... */
	{ 104690 },
	/* branches-loads-ops\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 104489 },
	/* branches-loads-reference\000legacy cache\000Branch prediction unit read accesses... */
	{ 104384 },
	/* branches-loads-refs\000legacy cache\000Branch prediction unit read accesses\000l... */
	{ 104284 },
	/* branches-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 106086 },
	/* branches-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 105986 },
	/* branches-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 105797 },
	/* branches-read\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 104900 },
	/* branches-read-access\000legacy cache\000Branch prediction unit read accesses\000... */
	{ 105295 },
	/* branches-read-miss\000legacy cache\000Branch prediction unit read misses\000lega... */
	{ 105501 },
	/* branches-read-misses\000legacy cache\000Branch prediction unit read misses\000le... */
	{ 105396 },
	/* branches-read-ops\000legacy cache\000Branch prediction unit read accesses\000leg... */
	{ 105197 },
	/* branches-read-reference\000legacy cache\000Branch prediction unit read accesses\... */
	{ 105093 },
	/* branches-read-refs\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 104994 },
	/* branches-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 105698 },
	/* branches-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 105604 },
	/* btb\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-conf... */
	{ 108832 },
	/* btb-access\000legacy cache\000Branch prediction unit read accesses\000legacy-cac... */
	{ 111201 },
	/* btb-load\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 108916 },
	/* btb-load-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 109291 },
	/* btb-load-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 109487 },
	/* btb-load-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 109387 },
	/* btb-load-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 109198 },
	/* btb-load-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 109099 },
	/* btb-load-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 109005 },
	/* btb-loads\000legacy cache\000Branch prediction unit read accesses\000legacy-cach... */
	{ 109585 },
	/* btb-loads-access\000legacy cache\000Branch prediction unit read accesses\000lega... */
	{ 109964 },
	/* btb-loads-miss\000legacy cache\000Branch prediction unit read misses\000legacy-c... */
	{ 110162 },
	/* btb-loads-misses\000legacy cache\000Branch prediction unit read misses\000legacy... */
	{ 110061 },
	/* btb-loads-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 109870 },
	/* btb-loads-reference\000legacy cache\000Branch prediction unit read accesses\000l... */
	{ 109770 },
	/* btb-loads-refs\000legacy cache\000Branch prediction unit read accesses\000legacy... */
	{ 109675 },
	/* btb-miss\000legacy cache\000Branch prediction unit read misses\000legacy-cache-c... */
	{ 111387 },
	/* btb-misses\000legacy cache\000Branch prediction unit read misses\000legacy-cache... */
	{ 111292 },
	/* btb-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-cache-... */
	{ 111113 },
	/* btb-read\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 110261 },
	/* btb-read-access\000legacy cache\000Branch prediction unit read accesses\000legac... */
	{ 110636 },
	/* btb-read-miss\000legacy cache\000Branch prediction unit read misses\000legacy-ca... */
	{ 110832 },
	/* btb-read-misses\000legacy cache\000Branch prediction unit read misses\000legacy-... */
	{ 110732 },
	/* btb-read-ops\000legacy cache\000Branch prediction unit read accesses\000legacy-c... */
	{ 110543 },
	/* btb-read-reference\000legacy cache\000Branch prediction unit read accesses\000le... */
	{ 110444 },
	/* btb-read-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 110350 },
	/* btb-reference\000legacy cache\000Branch prediction unit read accesses\000legacy-... */
	{ 111019 },
	/* btb-refs\000legacy cache\000Branch prediction unit read accesses\000legacy-cache... */
	{ 110930 },
	/* bus-cycles\000legacy hardware\000Bus cycles, which can be different from total c... */
	{ 122682 },
	/* cache-misses\000legacy hardware\000Cache misses. Usually this indicates Last Lev... */
	{ 122075 },
	/* cache-references\000legacy hardware\000Cache accesses. Usually this indicates La... */
	{ 121805 },
	/* cpu-cycles\000legacy hardware\000Total cycles. Be wary of what happens during CP... */
	{ 121305 },
	/* cycles\000legacy hardware\000Total cycles. Be wary of what happens during CPU fr... */
	{ 121467 },
	/* d-tlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\000... */
	{ 78952 },
	/* d-tlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\... */
	{ 85655 },
	/* d-tlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\00... */
	{ 79024 },
	/* d-tlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-conf... */
	{ 79351 },
	/* d-tlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0... */
	{ 79523 },
	/* d-tlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config... */
	{ 79435 },
	/* d-tlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 79270 },
	/* d-tlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-c... */
	{ 79183 },
	/* d-tlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 79101 },
	/* d-tlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\0... */
	{ 79609 },
	/* d-tlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-con... */
	{ 79940 },
	/* d-tlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=... */
	{ 80114 },
	/* d-tlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-confi... */
	{ 80025 },
	/* d-tlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 79858 },
	/* d-tlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-... */
	{ 79770 },
	/* d-tlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-confi... */
	{ 79687 },
	/* d-tlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x1000... */
	{ 85817 },
	/* d-tlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10... */
	{ 85734 },
	/* d-tlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000... */
	{ 85579 },
	/* d-tlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-con... */
	{ 82650 },
	/* d-tlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy-ca... */
	{ 83025 },
	/* d-tlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-... */
	{ 83217 },
	/* d-tlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cach... */
	{ 83121 },
	/* d-tlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache... */
	{ 82932 },
	/* d-tlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000legacy... */
	{ 82833 },
	/* d-tlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cach... */
	{ 82739 },
	/* d-tlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-c... */
	{ 83311 },
	/* d-tlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000legacy-... */
	{ 83694 },
	/* d-tlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cach... */
	{ 83890 },
	/* d-tlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy-ca... */
	{ 83792 },
	/* d-tlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cac... */
	{ 83599 },
	/* d-tlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 83498 },
	/* d-tlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-ca... */
	{ 83402 },
	/* d-tlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\00... */
	{ 80201 },
	/* d-tlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-conf... */
	{ 80528 },
	/* d-tlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0... */
	{ 80700 },
	/* d-tlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config... */
	{ 80612 },
	/* d-tlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 80447 },
	/* d-tlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-c... */
	{ 80360 },
	/* d-tlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 80278 },
	/* d-tlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 85497 },
	/* d-tlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\00... */
	{ 85420 },
	/* d-tlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legacy-c... */
	{ 84703 },
	/* d-tlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\000l... */
	{ 85110 },
	/* d-tlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000legac... */
	{ 85318 },
	/* d-tlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000leg... */
	{ 85214 },
	/* d-tlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 85009 },
	/* d-tlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesses\0... */
	{ 84902 },
	/* d-tlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000leg... */
	{ 84800 },
	/* d-tlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legacy-c... */
	{ 83986 },
	/* d-tlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\000l... */
	{ 84393 },
	/* d-tlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000legac... */
	{ 84601 },
	/* d-tlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000leg... */
	{ 84497 },
	/* d-tlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 84292 },
	/* d-tlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesses\0... */
	{ 84185 },
	/* d-tlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000leg... */
	{ 84083 },
	/* d-tlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x... */
	{ 80786 },
	/* d-tlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache-co... */
	{ 81137 },
	/* d-tlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config... */
	{ 81317 },
	/* d-tlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-conf... */
	{ 81227 },
	/* d-tlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 81050 },
	/* d-tlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-cache... */
	{ 80957 },
	/* d-tlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-conf... */
	{ 80869 },
	/* d-tlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0... */
	{ 81405 },
	/* d-tlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cache-c... */
	{ 81760 },
	/* d-tlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-confi... */
	{ 81942 },
	/* d-tlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-con... */
	{ 81851 },
	/* d-tlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-conf... */
	{ 81672 },
	/* d-tlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-cach... */
	{ 81578 },
	/* d-tlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-con... */
	{ 81489 },
	/* d-tlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x... */
	{ 82031 },
	/* d-tlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache-co... */
	{ 82382 },
	/* d-tlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config... */
	{ 82562 },
	/* d-tlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-conf... */
	{ 82472 },
	/* d-tlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 82295 },
	/* d-tlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-cache... */
	{ 82202 },
	/* d-tlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-conf... */
	{ 82114 },
	/* data-tlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\... */
	{ 85898 },
	/* data-tlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 92823 },
	/* data-tlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3... */
	{ 85973 },
	/* data-tlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-c... */
	{ 86312 },
	/* data-tlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-confi... */
	{ 86490 },
	/* data-tlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-con... */
	{ 86399 },
	/* data-tlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-conf... */
	{ 86228 },
	/* data-tlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cach... */
	{ 86138 },
	/* data-tlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-con... */
	{ 86053 },
	/* data-tlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 86579 },
	/* data-tlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-... */
	{ 86922 },
	/* data-tlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-conf... */
	{ 87102 },
	/* data-tlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-co... */
	{ 87010 },
	/* data-tlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-con... */
	{ 86837 },
	/* data-tlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cac... */
	{ 86746 },
	/* data-tlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-co... */
	{ 86660 },
	/* data-tlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x1... */
	{ 92991 },
	/* data-tlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0... */
	{ 92905 },
	/* data-tlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\... */
	{ 92744 },
	/* data-tlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-... */
	{ 89725 },
	/* data-tlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy... */
	{ 90112 },
	/* data-tlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cac... */
	{ 90310 },
	/* data-tlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-c... */
	{ 90211 },
	/* data-tlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-ca... */
	{ 90016 },
	/* data-tlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000leg... */
	{ 89914 },
	/* data-tlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-c... */
	{ 89817 },
	/* data-tlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cach... */
	{ 90407 },
	/* data-tlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 90802 },
	/* data-tlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-c... */
	{ 91004 },
	/* data-tlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy... */
	{ 90903 },
	/* data-tlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-... */
	{ 90704 },
	/* data-tlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000l... */
	{ 90600 },
	/* data-tlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy... */
	{ 90501 },
	/* data-tlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3... */
	{ 87192 },
	/* data-tlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-c... */
	{ 87531 },
	/* data-tlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-confi... */
	{ 87709 },
	/* data-tlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-con... */
	{ 87618 },
	/* data-tlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-conf... */
	{ 87447 },
	/* data-tlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cach... */
	{ 87357 },
	/* data-tlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-con... */
	{ 87272 },
	/* data-tlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-con... */
	{ 92659 },
	/* data-tlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3... */
	{ 92579 },
	/* data-tlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legac... */
	{ 91841 },
	/* data-tlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\0... */
	{ 92260 },
	/* data-tlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000le... */
	{ 92474 },
	/* data-tlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000... */
	{ 92367 },
	/* data-tlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000l... */
	{ 92156 },
	/* data-tlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesse... */
	{ 92046 },
	/* data-tlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000... */
	{ 91941 },
	/* data-tlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legac... */
	{ 91103 },
	/* data-tlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\0... */
	{ 91522 },
	/* data-tlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000le... */
	{ 91736 },
	/* data-tlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000... */
	{ 91629 },
	/* data-tlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000l... */
	{ 91418 },
	/* data-tlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesse... */
	{ 91308 },
	/* data-tlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000... */
	{ 91203 },
	/* data-tlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config... */
	{ 87798 },
	/* data-tlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache... */
	{ 88161 },
	/* data-tlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-con... */
	{ 88347 },
	/* data-tlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-c... */
	{ 88254 },
	/* data-tlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-co... */
	{ 88071 },
	/* data-tlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-ca... */
	{ 87975 },
	/* data-tlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-c... */
	{ 87884 },
	/* data-tlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 88438 },
	/* data-tlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cach... */
	{ 88805 },
	/* data-tlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-co... */
	{ 88993 },
	/* data-tlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-... */
	{ 88899 },
	/* data-tlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-c... */
	{ 88714 },
	/* data-tlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-c... */
	{ 88617 },
	/* data-tlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-... */
	{ 88525 },
	/* data-tlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config... */
	{ 89085 },
	/* data-tlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache... */
	{ 89448 },
	/* data-tlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-con... */
	{ 89634 },
	/* data-tlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-c... */
	{ 89541 },
	/* data-tlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-co... */
	{ 89358 },
	/* data-tlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-ca... */
	{ 89262 },
	/* data-tlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-c... */
	{ 89171 },
	/* dtlb\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\0001... */
	{ 72083 },
	/* dtlb-access\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\0... */
	{ 78712 },
	/* dtlb-load\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000... */
	{ 72154 },
	/* dtlb-load-access\000legacy cache\000Data TLB read accesses\000legacy-cache-confi... */
	{ 72477 },
	/* dtlb-load-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x... */
	{ 72647 },
	/* dtlb-load-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=... */
	{ 72560 },
	/* dtlb-load-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3... */
	{ 72397 },
	/* dtlb-load-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-co... */
	{ 72311 },
	/* dtlb-load-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 72230 },
	/* dtlb-loads\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\00... */
	{ 72732 },
	/* dtlb-loads-access\000legacy cache\000Data TLB read accesses\000legacy-cache-conf... */
	{ 73059 },
	/* dtlb-loads-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0... */
	{ 73231 },
	/* dtlb-loads-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config... */
	{ 73143 },
	/* dtlb-loads-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 72978 },
	/* dtlb-loads-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-c... */
	{ 72891 },
	/* dtlb-loads-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config... */
	{ 72809 },
	/* dtlb-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x10003... */
	{ 78872 },
	/* dtlb-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x100... */
	{ 78790 },
	/* dtlb-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000\... */
	{ 78637 },
	/* dtlb-prefetch\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-conf... */
	{ 75738 },
	/* dtlb-prefetch-access\000legacy cache\000Data TLB prefetch accesses\000legacy-cac... */
	{ 76109 },
	/* dtlb-prefetch-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache-c... */
	{ 76299 },
	/* dtlb-prefetch-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cache... */
	{ 76204 },
	/* dtlb-prefetch-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-... */
	{ 76017 },
	/* dtlb-prefetch-reference\000legacy cache\000Data TLB prefetch accesses\000legacy-... */
	{ 75919 },
	/* dtlb-prefetch-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cache... */
	{ 75826 },
	/* dtlb-prefetches\000legacy cache\000Data TLB prefetch accesses\000legacy-cache-co... */
	{ 76392 },
	/* dtlb-prefetches-access\000legacy cache\000Data TLB prefetch accesses\000legacy-c... */
	{ 76771 },
	/* dtlb-prefetches-miss\000legacy cache\000Data TLB prefetch misses\000legacy-cache... */
	{ 76965 },
	/* dtlb-prefetches-misses\000legacy cache\000Data TLB prefetch misses\000legacy-cac... */
	{ 76868 },
	/* dtlb-prefetches-ops\000legacy cache\000Data TLB prefetch accesses\000legacy-cach... */
	{ 76677 },
	/* dtlb-prefetches-reference\000legacy cache\000Data TLB prefetch accesses\000legac... */
	{ 76577 },
	/* dtlb-prefetches-refs\000legacy cache\000Data TLB prefetch accesses\000legacy-cac... */
	{ 76482 },
	/* dtlb-read\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000... */
	{ 73317 },
	/* dtlb-read-access\000legacy cache\000Data TLB read accesses\000legacy-cache-confi... */
	{ 73640 },
	/* dtlb-read-miss\000legacy cache\000Data TLB read misses\000legacy-cache-config=0x... */
	{ 73810 },
	/* dtlb-read-misses\000legacy cache\000Data TLB read misses\000legacy-cache-config=... */
	{ 73723 },
	/* dtlb-read-ops\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3... */
	{ 73560 },
	/* dtlb-read-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-co... */
	{ 73474 },
	/* dtlb-read-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 73393 },
	/* dtlb-reference\000legacy cache\000Data TLB read accesses\000legacy-cache-config=... */
	{ 78556 },
	/* dtlb-refs\000legacy cache\000Data TLB read accesses\000legacy-cache-config=3\000... */
	{ 78480 },
	/* dtlb-speculative-load\000legacy cache\000Data TLB prefetch accesses\000legacy-ca... */
	{ 77770 },
	/* dtlb-speculative-load-access\000legacy cache\000Data TLB prefetch accesses\000le... */
	{ 78173 },
	/* dtlb-speculative-load-miss\000legacy cache\000Data TLB prefetch misses\000legacy... */
	{ 78379 },
	/* dtlb-speculative-load-misses\000legacy cache\000Data TLB prefetch misses\000lega... */
	{ 78276 },
	/* dtlb-speculative-load-ops\000legacy cache\000Data TLB prefetch accesses\000legac... */
	{ 78073 },
	/* dtlb-speculative-load-reference\000legacy cache\000Data TLB prefetch accesses\00... */
	{ 77967 },
	/* dtlb-speculative-load-refs\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 77866 },
	/* dtlb-speculative-read\000legacy cache\000Data TLB prefetch accesses\000legacy-ca... */
	{ 77060 },
	/* dtlb-speculative-read-access\000legacy cache\000Data TLB prefetch accesses\000le... */
	{ 77463 },
	/* dtlb-speculative-read-miss\000legacy cache\000Data TLB prefetch misses\000legacy... */
	{ 77669 },
	/* dtlb-speculative-read-misses\000legacy cache\000Data TLB prefetch misses\000lega... */
	{ 77566 },
	/* dtlb-speculative-read-ops\000legacy cache\000Data TLB prefetch accesses\000legac... */
	{ 77363 },
	/* dtlb-speculative-read-reference\000legacy cache\000Data TLB prefetch accesses\00... */
	{ 77257 },
	/* dtlb-speculative-read-refs\000legacy cache\000Data TLB prefetch accesses\000lega... */
	{ 77156 },
	/* dtlb-store\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x1... */
	{ 73895 },
	/* dtlb-store-access\000legacy cache\000Data TLB write accesses\000legacy-cache-con... */
	{ 74242 },
	/* dtlb-store-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=... */
	{ 74420 },
	/* dtlb-store-misses\000legacy cache\000Data TLB write misses\000legacy-cache-confi... */
	{ 74331 },
	/* dtlb-store-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config... */
	{ 74156 },
	/* dtlb-store-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-... */
	{ 74064 },
	/* dtlb-store-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 73977 },
	/* dtlb-stores\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x... */
	{ 74507 },
	/* dtlb-stores-access\000legacy cache\000Data TLB write accesses\000legacy-cache-co... */
	{ 74858 },
	/* dtlb-stores-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config... */
	{ 75038 },
	/* dtlb-stores-misses\000legacy cache\000Data TLB write misses\000legacy-cache-conf... */
	{ 74948 },
	/* dtlb-stores-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 74771 },
	/* dtlb-stores-reference\000legacy cache\000Data TLB write accesses\000legacy-cache... */
	{ 74678 },
	/* dtlb-stores-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-conf... */
	{ 74590 },
	/* dtlb-write\000legacy cache\000Data TLB write accesses\000legacy-cache-config=0x1... */
	{ 75126 },
	/* dtlb-write-access\000legacy cache\000Data TLB write accesses\000legacy-cache-con... */
	{ 75473 },
	/* dtlb-write-miss\000legacy cache\000Data TLB write misses\000legacy-cache-config=... */
	{ 75651 },
	/* dtlb-write-misses\000legacy cache\000Data TLB write misses\000legacy-cache-confi... */
	{ 75562 },
	/* dtlb-write-ops\000legacy cache\000Data TLB write accesses\000legacy-cache-config... */
	{ 75387 },
	/* dtlb-write-reference\000legacy cache\000Data TLB write accesses\000legacy-cache-... */
	{ 75295 },
	/* dtlb-write-refs\000legacy cache\000Data TLB write accesses\000legacy-cache-confi... */
	{ 75208 },
	/* i-tlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\... */
	{ 95555 },
	/* i-tlb-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-co... */
	{ 97799 },
	/* i-tlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-cache-conf... */
	{ 95634 },
	/* i-tlb-load-access\000legacy cache\000Instruction TLB read accesses\000legacy-cac... */
	{ 95989 },
	/* i-tlb-load-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-c... */
	{ 96175 },
	/* i-tlb-load-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache... */
	{ 96080 },
	/* i-tlb-load-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 95901 },
	/* i-tlb-load-reference\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 95807 },
	/* i-tlb-load-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 95718 },
	/* i-tlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy-cache-con... */
	{ 96268 },
	/* i-tlb-loads-access\000legacy cache\000Instruction TLB read accesses\000legacy-ca... */
	{ 96627 },
	/* i-tlb-loads-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-... */
	{ 96815 },
	/* i-tlb-loads-misses\000legacy cache\000Instruction TLB read misses\000legacy-cach... */
	{ 96719 },
	/* i-tlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 96538 },
	/* i-tlb-loads-reference\000legacy cache\000Instruction TLB read accesses\000legacy... */
	{ 96443 },
	/* i-tlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cach... */
	{ 96353 },
	/* i-tlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config... */
	{ 97975 },
	/* i-tlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-conf... */
	{ 97885 },
	/* i-tlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-confi... */
	{ 97716 },
	/* i-tlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-cache-conf... */
	{ 96909 },
	/* i-tlb-read-access\000legacy cache\000Instruction TLB read accesses\000legacy-cac... */
	{ 97264 },
	/* i-tlb-read-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-c... */
	{ 97450 },
	/* i-tlb-read-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache... */
	{ 97355 },
	/* i-tlb-read-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 97176 },
	/* i-tlb-read-reference\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 97082 },
	/* i-tlb-read-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 96993 },
	/* i-tlb-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 97627 },
	/* i-tlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-conf... */
	{ 97543 },
	/* idle-cycles-backend\000legacy hardware\000Stalled cycles during retirement [This... */
	{ 123247 },
	/* idle-cycles-frontend\000legacy hardware\000Stalled cycles during issue [This eve... */
	{ 122945 },
	/* instruction-tlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 98063 },
	/* instruction-tlb-access\000legacy cache\000Instruction TLB read accesses\000legac... */
	{ 100557 },
	/* instruction-tlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 98152 },
	/* instruction-tlb-load-access\000legacy cache\000Instruction TLB read accesses\000... */
	{ 98547 },
	/* instruction-tlb-load-miss\000legacy cache\000Instruction TLB read misses\000lega... */
	{ 98753 },
	/* instruction-tlb-load-misses\000legacy cache\000Instruction TLB read misses\000le... */
	{ 98648 },
	/* instruction-tlb-load-ops\000legacy cache\000Instruction TLB read accesses\000leg... */
	{ 98449 },
	/* instruction-tlb-load-reference\000legacy cache\000Instruction TLB read accesses\... */
	{ 98345 },
	/* instruction-tlb-load-refs\000legacy cache\000Instruction TLB read accesses\000le... */
	{ 98246 },
	/* instruction-tlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy... */
	{ 98856 },
	/* instruction-tlb-loads-access\000legacy cache\000Instruction TLB read accesses\00... */
	{ 99255 },
	/* instruction-tlb-loads-miss\000legacy cache\000Instruction TLB read misses\000leg... */
	{ 99463 },
	/* instruction-tlb-loads-misses\000legacy cache\000Instruction TLB read misses\000l... */
	{ 99357 },
	/* instruction-tlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000le... */
	{ 99156 },
	/* instruction-tlb-loads-reference\000legacy cache\000Instruction TLB read accesses... */
	{ 99051 },
	/* instruction-tlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000l... */
	{ 98951 },
	/* instruction-tlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-ca... */
	{ 100753 },
	/* instruction-tlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-... */
	{ 100653 },
	/* instruction-tlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-c... */
	{ 100464 },
	/* instruction-tlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 99567 },
	/* instruction-tlb-read-access\000legacy cache\000Instruction TLB read accesses\000... */
	{ 99962 },
	/* instruction-tlb-read-miss\000legacy cache\000Instruction TLB read misses\000lega... */
	{ 100168 },
	/* instruction-tlb-read-misses\000legacy cache\000Instruction TLB read misses\000le... */
	{ 100063 },
	/* instruction-tlb-read-ops\000legacy cache\000Instruction TLB read accesses\000leg... */
	{ 99864 },
	/* instruction-tlb-read-reference\000legacy cache\000Instruction TLB read accesses\... */
	{ 99760 },
	/* instruction-tlb-read-refs\000legacy cache\000Instruction TLB read accesses\000le... */
	{ 99661 },
	/* instruction-tlb-reference\000legacy cache\000Instruction TLB read accesses\000le... */
	{ 100365 },
	/* instruction-tlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 100271 },
	/* instructions\000legacy hardware\000Retired instructions. Be careful, these can b... */
	{ 121629 },
	/* itlb\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config=4\0... */
	{ 93075 },
	/* itlb-access\000legacy cache\000Instruction TLB read accesses\000legacy-cache-con... */
	{ 95294 },
	/* itlb-load\000legacy cache\000Instruction TLB read accesses\000legacy-cache-confi... */
	{ 93153 },
	/* itlb-load-access\000legacy cache\000Instruction TLB read accesses\000legacy-cach... */
	{ 93504 },
	/* itlb-load-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-co... */
	{ 93688 },
	/* itlb-load-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-... */
	{ 93594 },
	/* itlb-load-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-c... */
	{ 93417 },
	/* itlb-load-reference\000legacy cache\000Instruction TLB read accesses\000legacy-c... */
	{ 93324 },
	/* itlb-load-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 93236 },
	/* itlb-loads\000legacy cache\000Instruction TLB read accesses\000legacy-cache-conf... */
	{ 93780 },
	/* itlb-loads-access\000legacy cache\000Instruction TLB read accesses\000legacy-cac... */
	{ 94135 },
	/* itlb-loads-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-c... */
	{ 94321 },
	/* itlb-loads-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache... */
	{ 94226 },
	/* itlb-loads-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 94047 },
	/* itlb-loads-reference\000legacy cache\000Instruction TLB read accesses\000legacy-... */
	{ 93953 },
	/* itlb-loads-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache... */
	{ 93864 },
	/* itlb-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-config=... */
	{ 95468 },
	/* itlb-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-confi... */
	{ 95379 },
	/* itlb-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-config... */
	{ 95212 },
	/* itlb-read\000legacy cache\000Instruction TLB read accesses\000legacy-cache-confi... */
	{ 94414 },
	/* itlb-read-access\000legacy cache\000Instruction TLB read accesses\000legacy-cach... */
	{ 94765 },
	/* itlb-read-miss\000legacy cache\000Instruction TLB read misses\000legacy-cache-co... */
	{ 94949 },
	/* itlb-read-misses\000legacy cache\000Instruction TLB read misses\000legacy-cache-... */
	{ 94855 },
	/* itlb-read-ops\000legacy cache\000Instruction TLB read accesses\000legacy-cache-c... */
	{ 94678 },
	/* itlb-read-reference\000legacy cache\000Instruction TLB read accesses\000legacy-c... */
	{ 94585 },
	/* itlb-read-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 94497 },
	/* itlb-reference\000legacy cache\000Instruction TLB read accesses\000legacy-cache-... */
	{ 95124 },
	/* itlb-refs\000legacy cache\000Instruction TLB read accesses\000legacy-cache-confi... */
	{ 95041 },
	/* l1-d\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=... */
	{ 8037 },
	/* l1-d-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-... */
	{ 15406 },
	/* l1-d-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-co... */
	{ 8118 },
	/* l1-d-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 8481 },
	/* l1-d-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache... */
	{ 8671 },
	/* l1-d-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cac... */
	{ 8574 },
	/* l1-d-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 8391 },
	/* l1-d-load-reference\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 8295 },
	/* l1-d-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 8204 },
	/* l1-d-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-c... */
	{ 8766 },
	/* l1-d-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-... */
	{ 9133 },
	/* l1-d-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cach... */
	{ 9325 },
	/* l1-d-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-ca... */
	{ 9227 },
	/* l1-d-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 9042 },
	/* l1-d-loads-reference\000legacy cache\000Level 1 data cache read accesses\000lega... */
	{ 8945 },
	/* l1-d-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-ca... */
	{ 8853 },
	/* l1-d-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-conf... */
	{ 15586 },
	/* l1-d-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-co... */
	{ 15494 },
	/* l1-d-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-con... */
	{ 15321 },
	/* l1-d-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-... */
	{ 12122 },
	/* l1-d-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 12533 },
	/* l1-d-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000lega... */
	{ 12743 },
	/* l1-d-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000le... */
	{ 12638 },
	/* l1-d-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000leg... */
	{ 12431 },
	/* l1-d-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\... */
	{ 12323 },
	/* l1-d-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000le... */
	{ 12220 },
	/* l1-d-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legac... */
	{ 12846 },
	/* l1-d-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\0... */
	{ 13265 },
	/* l1-d-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000le... */
	{ 13479 },
	/* l1-d-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000... */
	{ 13372 },
	/* l1-d-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000l... */
	{ 13161 },
	/* l1-d-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 13051 },
	/* l1-d-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 12946 },
	/* l1-d-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-co... */
	{ 9421 },
	/* l1-d-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 9784 },
	/* l1-d-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache... */
	{ 9974 },
	/* l1-d-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cac... */
	{ 9877 },
	/* l1-d-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 9694 },
	/* l1-d-read-reference\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 9598 },
	/* l1-d-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 9507 },
	/* l1-d-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 15230 },
	/* l1-d-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-co... */
	{ 15144 },
	/* l1-d-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\00... */
	{ 14364 },
	/* l1-d-speculative-load-access\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 14807 },
	/* l1-d-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses... */
	{ 15033 },
	/* l1-d-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch miss... */
	{ 14920 },
	/* l1-d-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 14697 },
	/* l1-d-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch a... */
	{ 14581 },
	/* l1-d-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch access... */
	{ 14470 },
	/* l1-d-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\00... */
	{ 13584 },
	/* l1-d-speculative-read-access\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 14027 },
	/* l1-d-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses... */
	{ 14253 },
	/* l1-d-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch miss... */
	{ 14140 },
	/* l1-d-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 13917 },
	/* l1-d-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch a... */
	{ 13801 },
	/* l1-d-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch access... */
	{ 13690 },
	/* l1-d-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-... */
	{ 10069 },
	/* l1-d-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy... */
	{ 10456 },
	/* l1-d-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cac... */
	{ 10654 },
	/* l1-d-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-c... */
	{ 10555 },
	/* l1-d-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 10360 },
	/* l1-d-store-reference\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 10258 },
	/* l1-d-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 10161 },
	/* l1-d-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache... */
	{ 10751 },
	/* l1-d-stores-access\000legacy cache\000Level 1 data cache write accesses\000legac... */
	{ 11142 },
	/* l1-d-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-ca... */
	{ 11342 },
	/* l1-d-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-... */
	{ 11242 },
	/* l1-d-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 11045 },
	/* l1-d-stores-reference\000legacy cache\000Level 1 data cache write accesses\000le... */
	{ 10942 },
	/* l1-d-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-... */
	{ 10844 },
	/* l1-d-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-... */
	{ 11440 },
	/* l1-d-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy... */
	{ 11827 },
	/* l1-d-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cac... */
	{ 12025 },
	/* l1-d-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-c... */
	{ 11926 },
	/* l1-d-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 11731 },
	/* l1-d-write-reference\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 11629 },
	/* l1-d-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 11532 },
	/* l1-data\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-conf... */
	{ 23238 },
	/* l1-data-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 30829 },
	/* l1-data-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache... */
	{ 23322 },
	/* l1-data-load-access\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 23697 },
	/* l1-data-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-ca... */
	{ 23893 },
	/* l1-data-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-... */
	{ 23793 },
	/* l1-data-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 23604 },
	/* l1-data-load-reference\000legacy cache\000Level 1 data cache read accesses\000le... */
	{ 23505 },
	/* l1-data-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-... */
	{ 23411 },
	/* l1-data-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 23991 },
	/* l1-data-loads-access\000legacy cache\000Level 1 data cache read accesses\000lega... */
	{ 24370 },
	/* l1-data-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-c... */
	{ 24568 },
	/* l1-data-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy... */
	{ 24467 },
	/* l1-data-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-... */
	{ 24276 },
	/* l1-data-loads-reference\000legacy cache\000Level 1 data cache read accesses\000l... */
	{ 24176 },
	/* l1-data-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy... */
	{ 24081 },
	/* l1-data-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-c... */
	{ 31015 },
	/* l1-data-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache... */
	{ 30920 },
	/* l1-data-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-... */
	{ 30741 },
	/* l1-data-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000lega... */
	{ 27452 },
	/* l1-data-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\... */
	{ 27875 },
	/* l1-data-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000l... */
	{ 28091 },
	/* l1-data-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\00... */
	{ 27983 },
	/* l1-data-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 27770 },
	/* l1-data-prefetch-reference\000legacy cache\000Level 1 data cache prefetch access... */
	{ 27659 },
	/* l1-data-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\00... */
	{ 27553 },
	/* l1-data-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000le... */
	{ 28197 },
	/* l1-data-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 28628 },
	/* l1-data-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\00... */
	{ 28848 },
	/* l1-data-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\... */
	{ 28738 },
	/* l1-data-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\0... */
	{ 28521 },
	/* l1-data-prefetches-reference\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 28408 },
	/* l1-data-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\... */
	{ 28300 },
	/* l1-data-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache... */
	{ 24667 },
	/* l1-data-read-access\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 25042 },
	/* l1-data-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-ca... */
	{ 25238 },
	/* l1-data-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-... */
	{ 25138 },
	/* l1-data-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 24949 },
	/* l1-data-read-reference\000legacy cache\000Level 1 data cache read accesses\000le... */
	{ 24850 },
	/* l1-data-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-... */
	{ 24756 },
	/* l1-data-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-... */
	{ 30647 },
	/* l1-data-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache... */
	{ 30558 },
	/* l1-data-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 29757 },
	/* l1-data-speculative-load-access\000legacy cache\000Level 1 data cache prefetch a... */
	{ 30212 },
	/* l1-data-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch mis... */
	{ 30444 },
	/* l1-data-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch m... */
	{ 30328 },
	/* l1-data-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 30099 },
	/* l1-data-speculative-load-reference\000legacy cache\000Level 1 data cache prefetc... */
	{ 29980 },
	/* l1-data-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch acc... */
	{ 29866 },
	/* l1-data-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 28956 },
	/* l1-data-speculative-read-access\000legacy cache\000Level 1 data cache prefetch a... */
	{ 29411 },
	/* l1-data-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch mis... */
	{ 29643 },
	/* l1-data-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch m... */
	{ 29527 },
	/* l1-data-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 29298 },
	/* l1-data-speculative-read-reference\000legacy cache\000Level 1 data cache prefetc... */
	{ 29179 },
	/* l1-data-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch acc... */
	{ 29065 },
	/* l1-data-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cac... */
	{ 25336 },
	/* l1-data-store-access\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 25735 },
	/* l1-data-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-... */
	{ 25939 },
	/* l1-data-store-misses\000legacy cache\000Level 1 data cache write misses\000legac... */
	{ 25837 },
	/* l1-data-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy... */
	{ 25636 },
	/* l1-data-store-reference\000legacy cache\000Level 1 data cache write accesses\000... */
	{ 25531 },
	/* l1-data-store-refs\000legacy cache\000Level 1 data cache write accesses\000legac... */
	{ 25431 },
	/* l1-data-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 26039 },
	/* l1-data-stores-access\000legacy cache\000Level 1 data cache write accesses\000le... */
	{ 26442 },
	/* l1-data-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy... */
	{ 26648 },
	/* l1-data-stores-misses\000legacy cache\000Level 1 data cache write misses\000lega... */
	{ 26545 },
	/* l1-data-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legac... */
	{ 26342 },
	/* l1-data-stores-reference\000legacy cache\000Level 1 data cache write accesses\00... */
	{ 26236 },
	/* l1-data-stores-refs\000legacy cache\000Level 1 data cache write accesses\000lega... */
	{ 26135 },
	/* l1-data-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cac... */
	{ 26749 },
	/* l1-data-write-access\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 27148 },
	/* l1-data-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-... */
	{ 27352 },
	/* l1-data-write-misses\000legacy cache\000Level 1 data cache write misses\000legac... */
	{ 27250 },
	/* l1-data-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy... */
	{ 27049 },
	/* l1-data-write-reference\000legacy cache\000Level 1 data cache write accesses\000... */
	{ 26944 },
	/* l1-data-write-refs\000legacy cache\000Level 1 data cache write accesses\000legac... */
	{ 26844 },
	/* l1-dcache\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-co... */
	{ 13 },
	/* l1-dcache-access\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 7752 },
	/* l1-dcache-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 99 },
	/* l1-dcache-load-access\000legacy cache\000Level 1 data cache read accesses\000leg... */
	{ 482 },
	/* l1-dcache-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-... */
	{ 682 },
	/* l1-dcache-load-misses\000legacy cache\000Level 1 data cache read misses\000legac... */
	{ 580 },
	/* l1-dcache-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy... */
	{ 387 },
	/* l1-dcache-load-reference\000legacy cache\000Level 1 data cache read accesses\000... */
	{ 286 },
	/* l1-dcache-load-refs\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 190 },
	/* l1-dcache-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-ca... */
	{ 782 },
	/* l1-dcache-loads-access\000legacy cache\000Level 1 data cache read accesses\000le... */
	{ 1169 },
	/* l1-dcache-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy... */
	{ 1371 },
	/* l1-dcache-loads-misses\000legacy cache\000Level 1 data cache read misses\000lega... */
	{ 1268 },
	/* l1-dcache-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 1073 },
	/* l1-dcache-loads-reference\000legacy cache\000Level 1 data cache read accesses\00... */
	{ 971 },
	/* l1-dcache-loads-refs\000legacy cache\000Level 1 data cache read accesses\000lega... */
	{ 874 },
	/* l1-dcache-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache... */
	{ 7942 },
	/* l1-dcache-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cac... */
	{ 7845 },
	/* l1-dcache-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 7662 },
	/* l1-dcache-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000le... */
	{ 4313 },
	/* l1-dcache-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 4744 },
	/* l1-dcache-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\00... */
	{ 4964 },
	/* l1-dcache-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\... */
	{ 4854 },
	/* l1-dcache-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\0... */
	{ 4637 },
	/* l1-dcache-prefetch-reference\000legacy cache\000Level 1 data cache prefetch acce... */
	{ 4524 },
	/* l1-dcache-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\... */
	{ 4416 },
	/* l1-dcache-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 5072 },
	/* l1-dcache-prefetches-access\000legacy cache\000Level 1 data cache prefetch acces... */
	{ 5511 },
	/* l1-dcache-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\... */
	{ 5735 },
	/* l1-dcache-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misse... */
	{ 5623 },
	/* l1-dcache-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 5402 },
	/* l1-dcache-prefetches-reference\000legacy cache\000Level 1 data cache prefetch ac... */
	{ 5287 },
	/* l1-dcache-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 5177 },
	/* l1-dcache-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 1472 },
	/* l1-dcache-read-access\000legacy cache\000Level 1 data cache read accesses\000leg... */
	{ 1855 },
	/* l1-dcache-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-... */
	{ 2055 },
	/* l1-dcache-read-misses\000legacy cache\000Level 1 data cache read misses\000legac... */
	{ 1953 },
	/* l1-dcache-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy... */
	{ 1760 },
	/* l1-dcache-read-reference\000legacy cache\000Level 1 data cache read accesses\000... */
	{ 1659 },
	/* l1-dcache-read-refs\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 1563 },
	/* l1-dcache-reference\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 7566 },
	/* l1-dcache-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 7475 },
	/* l1-dcache-speculative-load\000legacy cache\000Level 1 data cache prefetch access... */
	{ 6660 },
	/* l1-dcache-speculative-load-access\000legacy cache\000Level 1 data cache prefetch... */
	{ 7123 },
	/* l1-dcache-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch m... */
	{ 7359 },
	/* l1-dcache-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch... */
	{ 7241 },
	/* l1-dcache-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch ac... */
	{ 7008 },
	/* l1-dcache-speculative-load-reference\000legacy cache\000Level 1 data cache prefe... */
	{ 6887 },
	/* l1-dcache-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch a... */
	{ 6771 },
	/* l1-dcache-speculative-read\000legacy cache\000Level 1 data cache prefetch access... */
	{ 5845 },
	/* l1-dcache-speculative-read-access\000legacy cache\000Level 1 data cache prefetch... */
	{ 6308 },
	/* l1-dcache-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch m... */
	{ 6544 },
	/* l1-dcache-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch... */
	{ 6426 },
	/* l1-dcache-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch ac... */
	{ 6193 },
	/* l1-dcache-speculative-read-reference\000legacy cache\000Level 1 data cache prefe... */
	{ 6072 },
	/* l1-dcache-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch a... */
	{ 5956 },
	/* l1-dcache-store\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 2155 },
	/* l1-dcache-store-access\000legacy cache\000Level 1 data cache write accesses\000l... */
	{ 2562 },
	/* l1-dcache-store-miss\000legacy cache\000Level 1 data cache write misses\000legac... */
	{ 2770 },
	/* l1-dcache-store-misses\000legacy cache\000Level 1 data cache write misses\000leg... */
	{ 2666 },
	/* l1-dcache-store-ops\000legacy cache\000Level 1 data cache write accesses\000lega... */
	{ 2461 },
	/* l1-dcache-store-reference\000legacy cache\000Level 1 data cache write accesses\0... */
	{ 2354 },
	/* l1-dcache-store-refs\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 2252 },
	/* l1-dcache-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-... */
	{ 2872 },
	/* l1-dcache-stores-access\000legacy cache\000Level 1 data cache write accesses\000... */
	{ 3283 },
	/* l1-dcache-stores-miss\000legacy cache\000Level 1 data cache write misses\000lega... */
	{ 3493 },
	/* l1-dcache-stores-misses\000legacy cache\000Level 1 data cache write misses\000le... */
	{ 3388 },
	/* l1-dcache-stores-ops\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 3181 },
	/* l1-dcache-stores-reference\000legacy cache\000Level 1 data cache write accesses\... */
	{ 3073 },
	/* l1-dcache-stores-refs\000legacy cache\000Level 1 data cache write accesses\000le... */
	{ 2970 },
	/* l1-dcache-write\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 3596 },
	/* l1-dcache-write-access\000legacy cache\000Level 1 data cache write accesses\000l... */
	{ 4003 },
	/* l1-dcache-write-miss\000legacy cache\000Level 1 data cache write misses\000legac... */
	{ 4211 },
	/* l1-dcache-write-misses\000legacy cache\000Level 1 data cache write misses\000leg... */
	{ 4107 },
	/* l1-dcache-write-ops\000legacy cache\000Level 1 data cache write accesses\000lega... */
	{ 3902 },
	/* l1-dcache-write-reference\000legacy cache\000Level 1 data cache write accesses\0... */
	{ 3795 },
	/* l1-dcache-write-refs\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 3693 },
	/* l1-i\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-... */
	{ 37366 },
	/* l1-i-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy... */
	{ 43053 },
	/* l1-i-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-c... */
	{ 37454 },
	/* l1-i-load-access\000legacy cache\000Level 1 instruction cache read accesses\000l... */
	{ 37845 },
	/* l1-i-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legac... */
	{ 38049 },
	/* l1-i-load-misses\000legacy cache\000Level 1 instruction cache read misses\000leg... */
	{ 37945 },
	/* l1-i-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 37748 },
	/* l1-i-load-reference\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 37645 },
	/* l1-i-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 37547 },
	/* l1-i-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-... */
	{ 38151 },
	/* l1-i-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000... */
	{ 38546 },
	/* l1-i-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000lega... */
	{ 38752 },
	/* l1-i-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000le... */
	{ 38647 },
	/* l1-i-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 38448 },
	/* l1-i-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\... */
	{ 38344 },
	/* l1-i-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000le... */
	{ 38245 },
	/* l1-i-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cac... */
	{ 43247 },
	/* l1-i-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-c... */
	{ 43148 },
	/* l1-i-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-ca... */
	{ 42961 },
	/* l1-i-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000... */
	{ 39552 },
	/* l1-i-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch acces... */
	{ 39991 },
	/* l1-i-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\... */
	{ 40215 },
	/* l1-i-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misse... */
	{ 40103 },
	/* l1-i-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses... */
	{ 39882 },
	/* l1-i-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch ac... */
	{ 39767 },
	/* l1-i-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesse... */
	{ 39657 },
	/* l1-i-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\0... */
	{ 40325 },
	/* l1-i-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch acc... */
	{ 40772 },
	/* l1-i-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misse... */
	{ 41000 },
	/* l1-i-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch mis... */
	{ 40886 },
	/* l1-i-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch access... */
	{ 40661 },
	/* l1-i-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 40544 },
	/* l1-i-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch acces... */
	{ 40432 },
	/* l1-i-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-c... */
	{ 38855 },
	/* l1-i-read-access\000legacy cache\000Level 1 instruction cache read accesses\000l... */
	{ 39246 },
	/* l1-i-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legac... */
	{ 39450 },
	/* l1-i-read-misses\000legacy cache\000Level 1 instruction cache read misses\000leg... */
	{ 39346 },
	/* l1-i-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 39149 },
	/* l1-i-read-reference\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 39046 },
	/* l1-i-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 38948 },
	/* l1-i-reference\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 42863 },
	/* l1-i-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-c... */
	{ 42770 },
	/* l1-i-speculative-load\000legacy cache\000Level 1 instruction cache prefetch acce... */
	{ 41941 },
	/* l1-i-speculative-load-access\000legacy cache\000Level 1 instruction cache prefet... */
	{ 42412 },
	/* l1-i-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 42652 },
	/* l1-i-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefet... */
	{ 42532 },
	/* l1-i-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 42295 },
	/* l1-i-speculative-load-reference\000legacy cache\000Level 1 instruction cache pre... */
	{ 42172 },
	/* l1-i-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 42054 },
	/* l1-i-speculative-read\000legacy cache\000Level 1 instruction cache prefetch acce... */
	{ 41112 },
	/* l1-i-speculative-read-access\000legacy cache\000Level 1 instruction cache prefet... */
	{ 41583 },
	/* l1-i-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 41823 },
	/* l1-i-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefet... */
	{ 41703 },
	/* l1-i-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 41466 },
	/* l1-i-speculative-read-reference\000legacy cache\000Level 1 instruction cache pre... */
	{ 41343 },
	/* l1-i-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 41225 },
	/* l1-icache\000legacy cache\000Level 1 instruction cache read accesses\000legacy-c... */
	{ 31108 },
	/* l1-icache-access\000legacy cache\000Level 1 instruction cache read accesses\000l... */
	{ 37060 },
	/* l1-icache-load\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 31201 },
	/* l1-icache-load-access\000legacy cache\000Level 1 instruction cache read accesses... */
	{ 31612 },
	/* l1-icache-load-miss\000legacy cache\000Level 1 instruction cache read misses\000... */
	{ 31826 },
	/* l1-icache-load-misses\000legacy cache\000Level 1 instruction cache read misses\0... */
	{ 31717 },
	/* l1-icache-load-ops\000legacy cache\000Level 1 instruction cache read accesses\00... */
	{ 31510 },
	/* l1-icache-load-reference\000legacy cache\000Level 1 instruction cache read acces... */
	{ 31402 },
	/* l1-icache-load-refs\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 31299 },
	/* l1-icache-loads\000legacy cache\000Level 1 instruction cache read accesses\000le... */
	{ 31933 },
	/* l1-icache-loads-access\000legacy cache\000Level 1 instruction cache read accesse... */
	{ 32348 },
	/* l1-icache-loads-miss\000legacy cache\000Level 1 instruction cache read misses\00... */
	{ 32564 },
	/* l1-icache-loads-misses\000legacy cache\000Level 1 instruction cache read misses\... */
	{ 32454 },
	/* l1-icache-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 32245 },
	/* l1-icache-loads-reference\000legacy cache\000Level 1 instruction cache read acce... */
	{ 32136 },
	/* l1-icache-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\... */
	{ 32032 },
	/* l1-icache-miss\000legacy cache\000Level 1 instruction cache read misses\000legac... */
	{ 37264 },
	/* l1-icache-misses\000legacy cache\000Level 1 instruction cache read misses\000leg... */
	{ 37160 },
	/* l1-icache-ops\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 36963 },
	/* l1-icache-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesse... */
	{ 33404 },
	/* l1-icache-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 33863 },
	/* l1-icache-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch mi... */
	{ 34097 },
	/* l1-icache-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 33980 },
	/* l1-icache-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch acc... */
	{ 33749 },
	/* l1-icache-prefetch-reference\000legacy cache\000Level 1 instruction cache prefet... */
	{ 33629 },
	/* l1-icache-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch ac... */
	{ 33514 },
	/* l1-icache-prefetches\000legacy cache\000Level 1 instruction cache prefetch acces... */
	{ 34212 },
	/* l1-icache-prefetches-access\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 34679 },
	/* l1-icache-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 34917 },
	/* l1-icache-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 34798 },
	/* l1-icache-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch a... */
	{ 34563 },
	/* l1-icache-prefetches-reference\000legacy cache\000Level 1 instruction cache pref... */
	{ 34441 },
	/* l1-icache-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 34324 },
	/* l1-icache-read\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 32672 },
	/* l1-icache-read-access\000legacy cache\000Level 1 instruction cache read accesses... */
	{ 33083 },
	/* l1-icache-read-miss\000legacy cache\000Level 1 instruction cache read misses\000... */
	{ 33297 },
	/* l1-icache-read-misses\000legacy cache\000Level 1 instruction cache read misses\0... */
	{ 33188 },
	/* l1-icache-read-ops\000legacy cache\000Level 1 instruction cache read accesses\00... */
	{ 32981 },
	/* l1-icache-read-reference\000legacy cache\000Level 1 instruction cache read acces... */
	{ 32873 },
	/* l1-icache-read-refs\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 32770 },
	/* l1-icache-reference\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 36860 },
	/* l1-icache-refs\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 36762 },
	/* l1-icache-speculative-load\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 35898 },
	/* l1-icache-speculative-load-access\000legacy cache\000Level 1 instruction cache p... */
	{ 36389 },
	/* l1-icache-speculative-load-miss\000legacy cache\000Level 1 instruction cache pre... */
	{ 36639 },
	/* l1-icache-speculative-load-misses\000legacy cache\000Level 1 instruction cache p... */
	{ 36514 },
	/* l1-icache-speculative-load-ops\000legacy cache\000Level 1 instruction cache pref... */
	{ 36267 },
	/* l1-icache-speculative-load-reference\000legacy cache\000Level 1 instruction cach... */
	{ 36139 },
	/* l1-icache-speculative-load-refs\000legacy cache\000Level 1 instruction cache pre... */
	{ 36016 },
	/* l1-icache-speculative-read\000legacy cache\000Level 1 instruction cache prefetch... */
	{ 35034 },
	/* l1-icache-speculative-read-access\000legacy cache\000Level 1 instruction cache p... */
	{ 35525 },
	/* l1-icache-speculative-read-miss\000legacy cache\000Level 1 instruction cache pre... */
	{ 35775 },
	/* l1-icache-speculative-read-misses\000legacy cache\000Level 1 instruction cache p... */
	{ 35650 },
	/* l1-icache-speculative-read-ops\000legacy cache\000Level 1 instruction cache pref... */
	{ 35403 },
	/* l1-icache-speculative-read-reference\000legacy cache\000Level 1 instruction cach... */
	{ 35275 },
	/* l1-icache-speculative-read-refs\000legacy cache\000Level 1 instruction cache pre... */
	{ 35152 },
	/* l1-instruction\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 49266 },
	/* l1-instruction-access\000legacy cache\000Level 1 instruction cache read accesses... */
	{ 55483 },
	/* l1-instruction-load\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 49364 },
	/* l1-instruction-load-access\000legacy cache\000Level 1 instruction cache read acc... */
	{ 49795 },
	/* l1-instruction-load-miss\000legacy cache\000Level 1 instruction cache read misse... */
	{ 50019 },
	/* l1-instruction-load-misses\000legacy cache\000Level 1 instruction cache read mis... */
	{ 49905 },
	/* l1-instruction-load-ops\000legacy cache\000Level 1 instruction cache read access... */
	{ 49688 },
	/* l1-instruction-load-reference\000legacy cache\000Level 1 instruction cache read ... */
	{ 49575 },
	/* l1-instruction-load-refs\000legacy cache\000Level 1 instruction cache read acces... */
	{ 49467 },
	/* l1-instruction-loads\000legacy cache\000Level 1 instruction cache read accesses\... */
	{ 50131 },
	/* l1-instruction-loads-access\000legacy cache\000Level 1 instruction cache read ac... */
	{ 50566 },
	/* l1-instruction-loads-miss\000legacy cache\000Level 1 instruction cache read miss... */
	{ 50792 },
	/* l1-instruction-loads-misses\000legacy cache\000Level 1 instruction cache read mi... */
	{ 50677 },
	/* l1-instruction-loads-ops\000legacy cache\000Level 1 instruction cache read acces... */
	{ 50458 },
	/* l1-instruction-loads-reference\000legacy cache\000Level 1 instruction cache read... */
	{ 50344 },
	/* l1-instruction-loads-refs\000legacy cache\000Level 1 instruction cache read acce... */
	{ 50235 },
	/* l1-instruction-miss\000legacy cache\000Level 1 instruction cache read misses\000... */
	{ 55697 },
	/* l1-instruction-misses\000legacy cache\000Level 1 instruction cache read misses\0... */
	{ 55588 },
	/* l1-instruction-ops\000legacy cache\000Level 1 instruction cache read accesses\00... */
	{ 55381 },
	/* l1-instruction-prefetch\000legacy cache\000Level 1 instruction cache prefetch ac... */
	{ 51672 },
	/* l1-instruction-prefetch-access\000legacy cache\000Level 1 instruction cache pref... */
	{ 52151 },
	/* l1-instruction-prefetch-miss\000legacy cache\000Level 1 instruction cache prefet... */
	{ 52395 },
	/* l1-instruction-prefetch-misses\000legacy cache\000Level 1 instruction cache pref... */
	{ 52273 },
	/* l1-instruction-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 52032 },
	/* l1-instruction-prefetch-reference\000legacy cache\000Level 1 instruction cache p... */
	{ 51907 },
	/* l1-instruction-prefetch-refs\000legacy cache\000Level 1 instruction cache prefet... */
	{ 51787 },
	/* l1-instruction-prefetches\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 52515 },
	/* l1-instruction-prefetches-access\000legacy cache\000Level 1 instruction cache pr... */
	{ 53002 },
	/* l1-instruction-prefetches-miss\000legacy cache\000Level 1 instruction cache pref... */
	{ 53250 },
	/* l1-instruction-prefetches-misses\000legacy cache\000Level 1 instruction cache pr... */
	{ 53126 },
	/* l1-instruction-prefetches-ops\000legacy cache\000Level 1 instruction cache prefe... */
	{ 52881 },
	/* l1-instruction-prefetches-reference\000legacy cache\000Level 1 instruction cache... */
	{ 52754 },
	/* l1-instruction-prefetches-refs\000legacy cache\000Level 1 instruction cache pref... */
	{ 52632 },
	/* l1-instruction-read\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 50905 },
	/* l1-instruction-read-access\000legacy cache\000Level 1 instruction cache read acc... */
	{ 51336 },
	/* l1-instruction-read-miss\000legacy cache\000Level 1 instruction cache read misse... */
	{ 51560 },
	/* l1-instruction-read-misses\000legacy cache\000Level 1 instruction cache read mis... */
	{ 51446 },
	/* l1-instruction-read-ops\000legacy cache\000Level 1 instruction cache read access... */
	{ 51229 },
	/* l1-instruction-read-reference\000legacy cache\000Level 1 instruction cache read ... */
	{ 51116 },
	/* l1-instruction-read-refs\000legacy cache\000Level 1 instruction cache read acces... */
	{ 51008 },
	/* l1-instruction-reference\000legacy cache\000Level 1 instruction cache read acces... */
	{ 55273 },
	/* l1-instruction-refs\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 55170 },
	/* l1-instruction-speculative-load\000legacy cache\000Level 1 instruction cache pre... */
	{ 54271 },
	/* l1-instruction-speculative-load-access\000legacy cache\000Level 1 instruction ca... */
	{ 54782 },
	/* l1-instruction-speculative-load-miss\000legacy cache\000Level 1 instruction cach... */
	{ 55042 },
	/* l1-instruction-speculative-load-misses\000legacy cache\000Level 1 instruction ca... */
	{ 54912 },
	/* l1-instruction-speculative-load-ops\000legacy cache\000Level 1 instruction cache... */
	{ 54655 },
	/* l1-instruction-speculative-load-reference\000legacy cache\000Level 1 instruction... */
	{ 54522 },
	/* l1-instruction-speculative-load-refs\000legacy cache\000Level 1 instruction cach... */
	{ 54394 },
	/* l1-instruction-speculative-read\000legacy cache\000Level 1 instruction cache pre... */
	{ 53372 },
	/* l1-instruction-speculative-read-access\000legacy cache\000Level 1 instruction ca... */
	{ 53883 },
	/* l1-instruction-speculative-read-miss\000legacy cache\000Level 1 instruction cach... */
	{ 54143 },
	/* l1-instruction-speculative-read-misses\000legacy cache\000Level 1 instruction ca... */
	{ 54013 },
	/* l1-instruction-speculative-read-ops\000legacy cache\000Level 1 instruction cache... */
	{ 53756 },
	/* l1-instruction-speculative-read-reference\000legacy cache\000Level 1 instruction... */
	{ 53623 },
	/* l1-instruction-speculative-read-refs\000legacy cache\000Level 1 instruction cach... */
	{ 53495 },
	/* l1d\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-config=0... */
	{ 15676 },
	/* l1d-access\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-c... */
	{ 22971 },
	/* l1d-load\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-con... */
	{ 15756 },
	/* l1d-load-access\000legacy cache\000Level 1 data cache read accesses\000legacy-ca... */
	{ 16115 },
	/* l1d-load-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-... */
	{ 16303 },
	/* l1d-load-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cach... */
	{ 16207 },
	/* l1d-load-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache... */
	{ 16026 },
	/* l1d-load-reference\000legacy cache\000Level 1 data cache read accesses\000legacy... */
	{ 15931 },
	/* l1d-load-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 15841 },
	/* l1d-loads\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-co... */
	{ 16397 },
	/* l1d-loads-access\000legacy cache\000Level 1 data cache read accesses\000legacy-c... */
	{ 16760 },
	/* l1d-loads-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache... */
	{ 16950 },
	/* l1d-loads-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cac... */
	{ 16853 },
	/* l1d-loads-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 16670 },
	/* l1d-loads-reference\000legacy cache\000Level 1 data cache read accesses\000legac... */
	{ 16574 },
	/* l1d-loads-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cac... */
	{ 16483 },
	/* l1d-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-confi... */
	{ 23149 },
	/* l1d-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cache-con... */
	{ 23058 },
	/* l1d-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-conf... */
	{ 22887 },
	/* l1d-prefetch\000legacy cache\000Level 1 data cache prefetch accesses\000legacy-c... */
	{ 19718 },
	/* l1d-prefetch-access\000legacy cache\000Level 1 data cache prefetch accesses\000l... */
	{ 20125 },
	/* l1d-prefetch-miss\000legacy cache\000Level 1 data cache prefetch misses\000legac... */
	{ 20333 },
	/* l1d-prefetch-misses\000legacy cache\000Level 1 data cache prefetch misses\000leg... */
	{ 20229 },
	/* l1d-prefetch-ops\000legacy cache\000Level 1 data cache prefetch accesses\000lega... */
	{ 20024 },
	/* l1d-prefetch-reference\000legacy cache\000Level 1 data cache prefetch accesses\0... */
	{ 19917 },
	/* l1d-prefetch-refs\000legacy cache\000Level 1 data cache prefetch accesses\000leg... */
	{ 19815 },
	/* l1d-prefetches\000legacy cache\000Level 1 data cache prefetch accesses\000legacy... */
	{ 20435 },
	/* l1d-prefetches-access\000legacy cache\000Level 1 data cache prefetch accesses\00... */
	{ 20850 },
	/* l1d-prefetches-miss\000legacy cache\000Level 1 data cache prefetch misses\000leg... */
	{ 21062 },
	/* l1d-prefetches-misses\000legacy cache\000Level 1 data cache prefetch misses\000l... */
	{ 20956 },
	/* l1d-prefetches-ops\000legacy cache\000Level 1 data cache prefetch accesses\000le... */
	{ 20747 },
	/* l1d-prefetches-reference\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 20638 },
	/* l1d-prefetches-refs\000legacy cache\000Level 1 data cache prefetch accesses\000l... */
	{ 20534 },
	/* l1d-read\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-con... */
	{ 17045 },
	/* l1d-read-access\000legacy cache\000Level 1 data cache read accesses\000legacy-ca... */
	{ 17404 },
	/* l1d-read-miss\000legacy cache\000Level 1 data cache read misses\000legacy-cache-... */
	{ 17592 },
	/* l1d-read-misses\000legacy cache\000Level 1 data cache read misses\000legacy-cach... */
	{ 17496 },
	/* l1d-read-ops\000legacy cache\000Level 1 data cache read accesses\000legacy-cache... */
	{ 17315 },
	/* l1d-read-reference\000legacy cache\000Level 1 data cache read accesses\000legacy... */
	{ 17220 },
	/* l1d-read-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 17130 },
	/* l1d-reference\000legacy cache\000Level 1 data cache read accesses\000legacy-cach... */
	{ 22797 },
	/* l1d-refs\000legacy cache\000Level 1 data cache read accesses\000legacy-cache-con... */
	{ 22712 },
	/* l1d-speculative-load\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 21939 },
	/* l1d-speculative-load-access\000legacy cache\000Level 1 data cache prefetch acces... */
	{ 22378 },
	/* l1d-speculative-load-miss\000legacy cache\000Level 1 data cache prefetch misses\... */
	{ 22602 },
	/* l1d-speculative-load-misses\000legacy cache\000Level 1 data cache prefetch misse... */
	{ 22490 },
	/* l1d-speculative-load-ops\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 22269 },
	/* l1d-speculative-load-reference\000legacy cache\000Level 1 data cache prefetch ac... */
	{ 22154 },
	/* l1d-speculative-load-refs\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 22044 },
	/* l1d-speculative-read\000legacy cache\000Level 1 data cache prefetch accesses\000... */
	{ 21166 },
	/* l1d-speculative-read-access\000legacy cache\000Level 1 data cache prefetch acces... */
	{ 21605 },
	/* l1d-speculative-read-miss\000legacy cache\000Level 1 data cache prefetch misses\... */
	{ 21829 },
	/* l1d-speculative-read-misses\000legacy cache\000Level 1 data cache prefetch misse... */
	{ 21717 },
	/* l1d-speculative-read-ops\000legacy cache\000Level 1 data cache prefetch accesses... */
	{ 21496 },
	/* l1d-speculative-read-reference\000legacy cache\000Level 1 data cache prefetch ac... */
	{ 21381 },
	/* l1d-speculative-read-refs\000legacy cache\000Level 1 data cache prefetch accesse... */
	{ 21271 },
	/* l1d-store\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-c... */
	{ 17686 },
	/* l1d-store-access\000legacy cache\000Level 1 data cache write accesses\000legacy-... */
	{ 18069 },
	/* l1d-store-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cach... */
	{ 18265 },
	/* l1d-store-misses\000legacy cache\000Level 1 data cache write misses\000legacy-ca... */
	{ 18167 },
	/* l1d-store-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cac... */
	{ 17974 },
	/* l1d-store-reference\000legacy cache\000Level 1 data cache write accesses\000lega... */
	{ 17873 },
	/* l1d-store-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 17777 },
	/* l1d-stores\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-... */
	{ 18361 },
	/* l1d-stores-access\000legacy cache\000Level 1 data cache write accesses\000legacy... */
	{ 18748 },
	/* l1d-stores-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cac... */
	{ 18946 },
	/* l1d-stores-misses\000legacy cache\000Level 1 data cache write misses\000legacy-c... */
	{ 18847 },
	/* l1d-stores-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 18652 },
	/* l1d-stores-reference\000legacy cache\000Level 1 data cache write accesses\000leg... */
	{ 18550 },
	/* l1d-stores-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-c... */
	{ 18453 },
	/* l1d-write\000legacy cache\000Level 1 data cache write accesses\000legacy-cache-c... */
	{ 19043 },
	/* l1d-write-access\000legacy cache\000Level 1 data cache write accesses\000legacy-... */
	{ 19426 },
	/* l1d-write-miss\000legacy cache\000Level 1 data cache write misses\000legacy-cach... */
	{ 19622 },
	/* l1d-write-misses\000legacy cache\000Level 1 data cache write misses\000legacy-ca... */
	{ 19524 },
	/* l1d-write-ops\000legacy cache\000Level 1 data cache write accesses\000legacy-cac... */
	{ 19331 },
	/* l1d-write-reference\000legacy cache\000Level 1 data cache write accesses\000lega... */
	{ 19230 },
	/* l1d-write-refs\000legacy cache\000Level 1 data cache write accesses\000legacy-ca... */
	{ 19134 },
	/* l1i\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cache-c... */
	{ 43344 },
	/* l1i-access\000legacy cache\000Level 1 instruction cache read accesses\000legacy-... */
	{ 48978 },
	/* l1i-load\000legacy cache\000Level 1 instruction cache read accesses\000legacy-ca... */
	{ 43431 },
	/* l1i-load-access\000legacy cache\000Level 1 instruction cache read accesses\000le... */
	{ 43818 },
	/* l1i-load-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy... */
	{ 44020 },
	/* l1i-load-misses\000legacy cache\000Level 1 instruction cache read misses\000lega... */
	{ 43917 },
	/* l1i-load-ops\000legacy cache\000Level 1 instruction cache read accesses\000legac... */
	{ 43722 },
	/* l1i-load-reference\000legacy cache\000Level 1 instruction cache read accesses\00... */
	{ 43620 },
	/* l1i-load-refs\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 43523 },
	/* l1i-loads\000legacy cache\000Level 1 instruction cache read accesses\000legacy-c... */
	{ 44121 },
	/* l1i-loads-access\000legacy cache\000Level 1 instruction cache read accesses\000l... */
	{ 44512 },
	/* l1i-loads-miss\000legacy cache\000Level 1 instruction cache read misses\000legac... */
	{ 44716 },
	/* l1i-loads-misses\000legacy cache\000Level 1 instruction cache read misses\000leg... */
	{ 44612 },
	/* l1i-loads-ops\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 44415 },
	/* l1i-loads-reference\000legacy cache\000Level 1 instruction cache read accesses\0... */
	{ 44312 },
	/* l1i-loads-refs\000legacy cache\000Level 1 instruction cache read accesses\000leg... */
	{ 44214 },
	/* l1i-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy-cach... */
	{ 49170 },
	/* l1i-misses\000legacy cache\000Level 1 instruction cache read misses\000legacy-ca... */
	{ 49072 },
	/* l1i-ops\000legacy cache\000Level 1 instruction cache read accesses\000legacy-cac... */
	{ 48887 },
	/* l1i-prefetch\000legacy cache\000Level 1 instruction cache prefetch accesses\000l... */
	{ 45508 },
	/* l1i-prefetch-access\000legacy cache\000Level 1 instruction cache prefetch access... */
	{ 45943 },
	/* l1i-prefetch-miss\000legacy cache\000Level 1 instruction cache prefetch misses\0... */
	{ 46165 },
	/* l1i-prefetch-misses\000legacy cache\000Level 1 instruction cache prefetch misses... */
	{ 46054 },
	/* l1i-prefetch-ops\000legacy cache\000Level 1 instruction cache prefetch accesses\... */
	{ 45835 },
	/* l1i-prefetch-reference\000legacy cache\000Level 1 instruction cache prefetch acc... */
	{ 45721 },
	/* l1i-prefetch-refs\000legacy cache\000Level 1 instruction cache prefetch accesses... */
	{ 45612 },
	/* l1i-prefetches\000legacy cache\000Level 1 instruction cache prefetch accesses\00... */
	{ 46274 },
	/* l1i-prefetches-access\000legacy cache\000Level 1 instruction cache prefetch acce... */
	{ 46717 },
	/* l1i-prefetches-miss\000legacy cache\000Level 1 instruction cache prefetch misses... */
	{ 46943 },
	/* l1i-prefetches-misses\000legacy cache\000Level 1 instruction cache prefetch miss... */
	{ 46830 },
	/* l1i-prefetches-ops\000legacy cache\000Level 1 instruction cache prefetch accesse... */
	{ 46607 },
	/* l1i-prefetches-reference\000legacy cache\000Level 1 instruction cache prefetch a... */
	{ 46491 },
	/* l1i-prefetches-refs\000legacy cache\000Level 1 instruction cache prefetch access... */
	{ 46380 },
	/* l1i-read\000legacy cache\000Level 1 instruction cache read accesses\000legacy-ca... */
	{ 44818 },
	/* l1i-read-access\000legacy cache\000Level 1 instruction cache read accesses\000le... */
	{ 45205 },
	/* l1i-read-miss\000legacy cache\000Level 1 instruction cache read misses\000legacy... */
	{ 45407 },
	/* l1i-read-misses\000legacy cache\000Level 1 instruction cache read misses\000lega... */
	{ 45304 },
	/* l1i-read-ops\000legacy cache\000Level 1 instruction cache read accesses\000legac... */
	{ 45109 },
	/* l1i-read-reference\000legacy cache\000Level 1 instruction cache read accesses\00... */
	{ 45007 },
	/* l1i-read-refs\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 44910 },
	/* l1i-reference\000legacy cache\000Level 1 instruction cache read accesses\000lega... */
	{ 48790 },
	/* l1i-refs\000legacy cache\000Level 1 instruction cache read accesses\000legacy-ca... */
	{ 48698 },
	/* l1i-speculative-load\000legacy cache\000Level 1 instruction cache prefetch acces... */
	{ 47876 },
	/* l1i-speculative-load-access\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 48343 },
	/* l1i-speculative-load-miss\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 48581 },
	/* l1i-speculative-load-misses\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 48462 },
	/* l1i-speculative-load-ops\000legacy cache\000Level 1 instruction cache prefetch a... */
	{ 48227 },
	/* l1i-speculative-load-reference\000legacy cache\000Level 1 instruction cache pref... */
	{ 48105 },
	/* l1i-speculative-load-refs\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 47988 },
	/* l1i-speculative-read\000legacy cache\000Level 1 instruction cache prefetch acces... */
	{ 47054 },
	/* l1i-speculative-read-access\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 47521 },
	/* l1i-speculative-read-miss\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 47759 },
	/* l1i-speculative-read-misses\000legacy cache\000Level 1 instruction cache prefetc... */
	{ 47640 },
	/* l1i-speculative-read-ops\000legacy cache\000Level 1 instruction cache prefetch a... */
	{ 47405 },
	/* l1i-speculative-read-reference\000legacy cache\000Level 1 instruction cache pref... */
	{ 47283 },
	/* l1i-speculative-read-refs\000legacy cache\000Level 1 instruction cache prefetch ... */
	{ 47166 },
	/* l2\000legacy cache\000Level 2 (or higher) last level cache read accesses\000lega... */
	{ 63212 },
	/* l2-access\000legacy cache\000Level 2 (or higher) last level cache read accesses\... */
	{ 71765 },
	/* l2-load\000legacy cache\000Level 2 (or higher) last level cache read accesses\00... */
	{ 63309 },
	/* l2-load-access\000legacy cache\000Level 2 (or higher) last level cache read acce... */
	{ 63736 },
	/* l2-load-miss\000legacy cache\000Level 2 (or higher) last level cache read misses... */
	{ 63958 },
	/* l2-load-misses\000legacy cache\000Level 2 (or higher) last level cache read miss... */
	{ 63845 },
	/* l2-load-ops\000legacy cache\000Level 2 (or higher) last level cache read accesse... */
	{ 63630 },
	/* l2-load-reference\000legacy cache\000Level 2 (or higher) last level cache read a... */
	{ 63518 },
	/* l2-load-refs\000legacy cache\000Level 2 (or higher) last level cache read access... */
	{ 63411 },
	/* l2-loads\000legacy cache\000Level 2 (or higher) last level cache read accesses\0... */
	{ 64069 },
	/* l2-loads-access\000legacy cache\000Level 2 (or higher) last level cache read acc... */
	{ 64500 },
	/* l2-loads-miss\000legacy cache\000Level 2 (or higher) last level cache read misse... */
	{ 64724 },
	/* l2-loads-misses\000legacy cache\000Level 2 (or higher) last level cache read mis... */
	{ 64610 },
	/* l2-loads-ops\000legacy cache\000Level 2 (or higher) last level cache read access... */
	{ 64393 },
	/* l2-loads-reference\000legacy cache\000Level 2 (or higher) last level cache read ... */
	{ 64280 },
	/* l2-loads-refs\000legacy cache\000Level 2 (or higher) last level cache read acces... */
	{ 64172 },
	/* l2-miss\000legacy cache\000Level 2 (or higher) last level cache read misses\000l... */
	{ 71977 },
	/* l2-misses\000legacy cache\000Level 2 (or higher) last level cache read misses\00... */
	{ 71869 },
	/* l2-ops\000legacy cache\000Level 2 (or higher) last level cache read accesses\000... */
	{ 71664 },
	/* l2-prefetch\000legacy cache\000Level 2 (or higher) last level cache prefetch acc... */
	{ 67985 },
	/* l2-prefetch-access\000legacy cache\000Level 2 (or higher) last level cache prefe... */
	{ 68460 },
	/* l2-prefetch-miss\000legacy cache\000Level 2 (or higher) last level cache prefetc... */
	{ 68702 },
	/* l2-prefetch-misses\000legacy cache\000Level 2 (or higher) last level cache prefe... */
	{ 68581 },
	/* l2-prefetch-ops\000legacy cache\000Level 2 (or higher) last level cache prefetch... */
	{ 68342 },
	/* l2-prefetch-reference\000legacy cache\000Level 2 (or higher) last level cache pr... */
	{ 68218 },
	/* l2-prefetch-refs\000legacy cache\000Level 2 (or higher) last level cache prefetc... */
	{ 68099 },
	/* l2-prefetches\000legacy cache\000Level 2 (or higher) last level cache prefetch a... */
	{ 68821 },
	/* l2-prefetches-access\000legacy cache\000Level 2 (or higher) last level cache pre... */
	{ 69304 },
	/* l2-prefetches-miss\000legacy cache\000Level 2 (or higher) last level cache prefe... */
	{ 69550 },
	/* l2-prefetches-misses\000legacy cache\000Level 2 (or higher) last level cache pre... */
	{ 69427 },
	/* l2-prefetches-ops\000legacy cache\000Level 2 (or higher) last level cache prefet... */
	{ 69184 },
	/* l2-prefetches-reference\000legacy cache\000Level 2 (or higher) last level cache ... */
	{ 69058 },
	/* l2-prefetches-refs\000legacy cache\000Level 2 (or higher) last level cache prefe... */
	{ 68937 },
	/* l2-read\000legacy cache\000Level 2 (or higher) last level cache read accesses\00... */
	{ 64836 },
	/* l2-read-access\000legacy cache\000Level 2 (or higher) last level cache read acce... */
	{ 65263 },
	/* l2-read-miss\000legacy cache\000Level 2 (or higher) last level cache read misses... */
	{ 65485 },
	/* l2-read-misses\000legacy cache\000Level 2 (or higher) last level cache read miss... */
	{ 65372 },
	/* l2-read-ops\000legacy cache\000Level 2 (or higher) last level cache read accesse... */
	{ 65157 },
	/* l2-read-reference\000legacy cache\000Level 2 (or higher) last level cache read a... */
	{ 65045 },
	/* l2-read-refs\000legacy cache\000Level 2 (or higher) last level cache read access... */
	{ 64938 },
	/* l2-reference\000legacy cache\000Level 2 (or higher) last level cache read access... */
	{ 71557 },
	/* l2-refs\000legacy cache\000Level 2 (or higher) last level cache read accesses\00... */
	{ 71455 },
	/* l2-speculative-load\000legacy cache\000Level 2 (or higher) last level cache pref... */
	{ 70563 },
	/* l2-speculative-load-access\000legacy cache\000Level 2 (or higher) last level cac... */
	{ 71070 },
	/* l2-speculative-load-miss\000legacy cache\000Level 2 (or higher) last level cache... */
	{ 71328 },
	/* l2-speculative-load-misses\000legacy cache\000Level 2 (or higher) last level cac... */
	{ 71199 },
	/* l2-speculative-load-ops\000legacy cache\000Level 2 (or higher) last level cache ... */
	{ 70944 },
	/* l2-speculative-load-reference\000legacy cache\000Level 2 (or higher) last level ... */
	{ 70812 },
	/* l2-speculative-load-refs\000legacy cache\000Level 2 (or higher) last level cache... */
	{ 70685 },
	/* l2-speculative-read\000legacy cache\000Level 2 (or higher) last level cache pref... */
	{ 69671 },
	/* l2-speculative-read-access\000legacy cache\000Level 2 (or higher) last level cac... */
	{ 70178 },
	/* l2-speculative-read-miss\000legacy cache\000Level 2 (or higher) last level cache... */
	{ 70436 },
	/* l2-speculative-read-misses\000legacy cache\000Level 2 (or higher) last level cac... */
	{ 70307 },
	/* l2-speculative-read-ops\000legacy cache\000Level 2 (or higher) last level cache ... */
	{ 70052 },
	/* l2-speculative-read-reference\000legacy cache\000Level 2 (or higher) last level ... */
	{ 69920 },
	/* l2-speculative-read-refs\000legacy cache\000Level 2 (or higher) last level cache... */
	{ 69793 },
	/* l2-store\000legacy cache\000Level 2 (or higher) last level cache write accesses\... */
	{ 65596 },
	/* l2-store-access\000legacy cache\000Level 2 (or higher) last level cache write ac... */
	{ 66047 },
	/* l2-store-miss\000legacy cache\000Level 2 (or higher) last level cache write miss... */
	{ 66277 },
	/* l2-store-misses\000legacy cache\000Level 2 (or higher) last level cache write mi... */
	{ 66162 },
	/* l2-store-ops\000legacy cache\000Level 2 (or higher) last level cache write acces... */
	{ 65935 },
	/* l2-store-reference\000legacy cache\000Level 2 (or higher) last level cache write... */
	{ 65817 },
	/* l2-store-refs\000legacy cache\000Level 2 (or higher) last level cache write acce... */
	{ 65704 },
	/* l2-stores\000legacy cache\000Level 2 (or higher) last level cache write accesses... */
	{ 66390 },
	/* l2-stores-access\000legacy cache\000Level 2 (or higher) last level cache write a... */
	{ 66845 },
	/* l2-stores-miss\000legacy cache\000Level 2 (or higher) last level cache write mis... */
	{ 67077 },
	/* l2-stores-misses\000legacy cache\000Level 2 (or higher) last level cache write m... */
	{ 66961 },
	/* l2-stores-ops\000legacy cache\000Level 2 (or higher) last level cache write acce... */
	{ 66732 },
	/* l2-stores-reference\000legacy cache\000Level 2 (or higher) last level cache writ... */
	{ 66613 },
	/* l2-stores-refs\000legacy cache\000Level 2 (or higher) last level cache write acc... */
	{ 66499 },
	/* l2-write\000legacy cache\000Level 2 (or higher) last level cache write accesses\... */
	{ 67191 },
	/* l2-write-access\000legacy cache\000Level 2 (or higher) last level cache write ac... */
	{ 67642 },
	/* l2-write-miss\000legacy cache\000Level 2 (or higher) last level cache write miss... */
	{ 67872 },
	/* l2-write-misses\000legacy cache\000Level 2 (or higher) last level cache write mi... */
	{ 67757 },
	/* l2-write-ops\000legacy cache\000Level 2 (or higher) last level cache write acces... */
	{ 67530 },
	/* l2-write-reference\000legacy cache\000Level 2 (or higher) last level cache write... */
	{ 67412 },
	/* l2-write-refs\000legacy cache\000Level 2 (or higher) last level cache write acce... */
	{ 67299 },
	/* llc\000legacy cache\000Last level cache read accesses\000legacy-cache-config=2\0... */
	{ 55804 },
	/* llc-access\000legacy cache\000Last level cache read accesses\000legacy-cache-con... */
	{ 62951 },
	/* llc-load\000legacy cache\000Last level cache read accesses\000legacy-cache-confi... */
	{ 55882 },
	/* llc-load-access\000legacy cache\000Last level cache read accesses\000legacy-cach... */
	{ 56233 },
	/* llc-load-miss\000legacy cache\000Last level cache read misses\000legacy-cache-co... */
	{ 56417 },
	/* llc-load-misses\000legacy cache\000Last level cache read misses\000legacy-cache-... */
	{ 56323 },
	/* llc-load-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-c... */
	{ 56146 },
	/* llc-load-reference\000legacy cache\000Last level cache read accesses\000legacy-c... */
	{ 56053 },
	/* llc-load-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-... */
	{ 55965 },
	/* llc-loads\000legacy cache\000Last level cache read accesses\000legacy-cache-conf... */
	{ 56509 },
	/* llc-loads-access\000legacy cache\000Last level cache read accesses\000legacy-cac... */
	{ 56864 },
	/* llc-loads-miss\000legacy cache\000Last level cache read misses\000legacy-cache-c... */
	{ 57050 },
	/* llc-loads-misses\000legacy cache\000Last level cache read misses\000legacy-cache... */
	{ 56955 },
	/* llc-loads-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-... */
	{ 56776 },
	/* llc-loads-reference\000legacy cache\000Last level cache read accesses\000legacy-... */
	{ 56682 },
	/* llc-loads-refs\000legacy cache\000Last level cache read accesses\000legacy-cache... */
	{ 56593 },
	/* llc-miss\000legacy cache\000Last level cache read misses\000legacy-cache-config=... */
	{ 63125 },
	/* llc-misses\000legacy cache\000Last level cache read misses\000legacy-cache-confi... */
	{ 63036 },
	/* llc-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-config... */
	{ 62869 },
	/* llc-prefetch\000legacy cache\000Last level cache prefetch accesses\000legacy-cac... */
	{ 59760 },
	/* llc-prefetch-access\000legacy cache\000Last level cache prefetch accesses\000leg... */
	{ 60159 },
	/* llc-prefetch-miss\000legacy cache\000Last level cache prefetch misses\000legacy-... */
	{ 60363 },
	/* llc-prefetch-misses\000legacy cache\000Last level cache prefetch misses\000legac... */
	{ 60261 },
	/* llc-prefetch-ops\000legacy cache\000Last level cache prefetch accesses\000legacy... */
	{ 60060 },
	/* llc-prefetch-reference\000legacy cache\000Last level cache prefetch accesses\000... */
	{ 59955 },
	/* llc-prefetch-refs\000legacy cache\000Last level cache prefetch accesses\000legac... */
	{ 59855 },
	/* llc-prefetches\000legacy cache\000Last level cache prefetch accesses\000legacy-c... */
	{ 60463 },
	/* llc-prefetches-access\000legacy cache\000Last level cache prefetch accesses\000l... */
	{ 60870 },
	/* llc-prefetches-miss\000legacy cache\000Last level cache prefetch misses\000legac... */
	{ 61078 },
	/* llc-prefetches-misses\000legacy cache\000Last level cache prefetch misses\000leg... */
	{ 60974 },
	/* llc-prefetches-ops\000legacy cache\000Last level cache prefetch accesses\000lega... */
	{ 60769 },
	/* llc-prefetches-reference\000legacy cache\000Last level cache prefetch accesses\0... */
	{ 60662 },
	/* llc-prefetches-refs\000legacy cache\000Last level cache prefetch accesses\000leg... */
	{ 60560 },
	/* llc-read\000legacy cache\000Last level cache read accesses\000legacy-cache-confi... */
	{ 57143 },
	/* llc-read-access\000legacy cache\000Last level cache read accesses\000legacy-cach... */
	{ 57494 },
	/* llc-read-miss\000legacy cache\000Last level cache read misses\000legacy-cache-co... */
	{ 57678 },
	/* llc-read-misses\000legacy cache\000Last level cache read misses\000legacy-cache-... */
	{ 57584 },
	/* llc-read-ops\000legacy cache\000Last level cache read accesses\000legacy-cache-c... */
	{ 57407 },
	/* llc-read-reference\000legacy cache\000Last level cache read accesses\000legacy-c... */
	{ 57314 },
	/* llc-read-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-... */
	{ 57226 },
	/* llc-reference\000legacy cache\000Last level cache read accesses\000legacy-cache-... */
	{ 62781 },
	/* llc-refs\000legacy cache\000Last level cache read accesses\000legacy-cache-confi... */
	{ 62698 },
	/* llc-speculative-load\000legacy cache\000Last level cache prefetch accesses\000le... */
	{ 61939 },
	/* llc-speculative-load-access\000legacy cache\000Last level cache prefetch accesse... */
	{ 62370 },
	/* llc-speculative-load-miss\000legacy cache\000Last level cache prefetch misses\00... */
	{ 62590 },
	/* llc-speculative-load-misses\000legacy cache\000Last level cache prefetch misses\... */
	{ 62480 },
	/* llc-speculative-load-ops\000legacy cache\000Last level cache prefetch accesses\0... */
	{ 62263 },
	/* llc-speculative-load-reference\000legacy cache\000Last level cache prefetch acce... */
	{ 62150 },
	/* llc-speculative-load-refs\000legacy cache\000Last level cache prefetch accesses\... */
	{ 62042 },
	/* llc-speculative-read\000legacy cache\000Last level cache prefetch accesses\000le... */
	{ 61180 },
	/* llc-speculative-read-access\000legacy cache\000Last level cache prefetch accesse... */
	{ 61611 },
	/* llc-speculative-read-miss\000legacy cache\000Last level cache prefetch misses\00... */
	{ 61831 },
	/* llc-speculative-read-misses\000legacy cache\000Last level cache prefetch misses\... */
	{ 61721 },
	/* llc-speculative-read-ops\000legacy cache\000Last level cache prefetch accesses\0... */
	{ 61504 },
	/* llc-speculative-read-reference\000legacy cache\000Last level cache prefetch acce... */
	{ 61391 },
	/* llc-speculative-read-refs\000legacy cache\000Last level cache prefetch accesses\... */
	{ 61283 },
	/* llc-store\000legacy cache\000Last level cache write accesses\000legacy-cache-con... */
	{ 57770 },
	/* llc-store-access\000legacy cache\000Last level cache write accesses\000legacy-ca... */
	{ 58145 },
	/* llc-store-miss\000legacy cache\000Last level cache write misses\000legacy-cache-... */
	{ 58337 },
	/* llc-store-misses\000legacy cache\000Last level cache write misses\000legacy-cach... */
	{ 58241 },
	/* llc-store-ops\000legacy cache\000Last level cache write accesses\000legacy-cache... */
	{ 58052 },
	/* llc-store-reference\000legacy cache\000Last level cache write accesses\000legacy... */
	{ 57953 },
	/* llc-store-refs\000legacy cache\000Last level cache write accesses\000legacy-cach... */
	{ 57859 },
	/* llc-stores\000legacy cache\000Last level cache write accesses\000legacy-cache-co... */
	{ 58431 },
	/* llc-stores-access\000legacy cache\000Last level cache write accesses\000legacy-c... */
	{ 58810 },
	/* llc-stores-miss\000legacy cache\000Last level cache write misses\000legacy-cache... */
	{ 59004 },
	/* llc-stores-misses\000legacy cache\000Last level cache write misses\000legacy-cac... */
	{ 58907 },
	/* llc-stores-ops\000legacy cache\000Last level cache write accesses\000legacy-cach... */
	{ 58716 },
	/* llc-stores-reference\000legacy cache\000Last level cache write accesses\000legac... */
	{ 58616 },
	/* llc-stores-refs\000legacy cache\000Last level cache write accesses\000legacy-cac... */
	{ 58521 },
	/* llc-write\000legacy cache\000Last level cache write accesses\000legacy-cache-con... */
	{ 59099 },
	/* llc-write-access\000legacy cache\000Last level cache write accesses\000legacy-ca... */
	{ 59474 },
	/* llc-write-miss\000legacy cache\000Last level cache write misses\000legacy-cache-... */
	{ 59666 },
	/* llc-write-misses\000legacy cache\000Last level cache write misses\000legacy-cach... */
	{ 59570 },
	/* llc-write-ops\000legacy cache\000Last level cache write accesses\000legacy-cache... */
	{ 59381 },
	/* llc-write-reference\000legacy cache\000Last level cache write accesses\000legacy... */
	{ 59282 },
	/* llc-write-refs\000legacy cache\000Last level cache write accesses\000legacy-cach... */
	{ 59188 },
	/* node\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\000\... */
	{ 114128 },
	/* node-access\000legacy cache\000Local memory read accesses\000legacy-cache-config... */
	{ 121053 },
	/* node-load\000legacy cache\000Local memory read accesses\000legacy-cache-config=6... */
	{ 114203 },
	/* node-load-access\000legacy cache\000Local memory read accesses\000legacy-cache-c... */
	{ 114542 },
	/* node-load-miss\000legacy cache\000Local memory read misses\000legacy-cache-confi... */
	{ 114720 },
	/* node-load-misses\000legacy cache\000Local memory read misses\000legacy-cache-con... */
	{ 114629 },
	/* node-load-ops\000legacy cache\000Local memory read accesses\000legacy-cache-conf... */
	{ 114458 },
	/* node-load-reference\000legacy cache\000Local memory read accesses\000legacy-cach... */
	{ 114368 },
	/* node-load-refs\000legacy cache\000Local memory read accesses\000legacy-cache-con... */
	{ 114283 },
	/* node-loads\000legacy cache\000Local memory read accesses\000legacy-cache-config=... */
	{ 114809 },
	/* node-loads-access\000legacy cache\000Local memory read accesses\000legacy-cache-... */
	{ 115152 },
	/* node-loads-miss\000legacy cache\000Local memory read misses\000legacy-cache-conf... */
	{ 115332 },
	/* node-loads-misses\000legacy cache\000Local memory read misses\000legacy-cache-co... */
	{ 115240 },
	/* node-loads-ops\000legacy cache\000Local memory read accesses\000legacy-cache-con... */
	{ 115067 },
	/* node-loads-reference\000legacy cache\000Local memory read accesses\000legacy-cac... */
	{ 114976 },
	/* node-loads-refs\000legacy cache\000Local memory read accesses\000legacy-cache-co... */
	{ 114890 },
	/* node-miss\000legacy cache\000Local memory read misses\000legacy-cache-config=0x1... */
	{ 121221 },
	/* node-misses\000legacy cache\000Local memory read misses\000legacy-cache-config=0... */
	{ 121135 },
	/* node-ops\000legacy cache\000Local memory read accesses\000legacy-cache-config=6\... */
	{ 120974 },
	/* node-prefetch\000legacy cache\000Local memory prefetch accesses\000legacy-cache-... */
	{ 117955 },
	/* node-prefetch-access\000legacy cache\000Local memory prefetch accesses\000legacy... */
	{ 118342 },
	/* node-prefetch-miss\000legacy cache\000Local memory prefetch misses\000legacy-cac... */
	{ 118540 },
	/* node-prefetch-misses\000legacy cache\000Local memory prefetch misses\000legacy-c... */
	{ 118441 },
	/* node-prefetch-ops\000legacy cache\000Local memory prefetch accesses\000legacy-ca... */
	{ 118246 },
	/* node-prefetch-reference\000legacy cache\000Local memory prefetch accesses\000leg... */
	{ 118144 },
	/* node-prefetch-refs\000legacy cache\000Local memory prefetch accesses\000legacy-c... */
	{ 118047 },
	/* node-prefetches\000legacy cache\000Local memory prefetch accesses\000legacy-cach... */
	{ 118637 },
	/* node-prefetches-access\000legacy cache\000Local memory prefetch accesses\000lega... */
	{ 119032 },
	/* node-prefetches-miss\000legacy cache\000Local memory prefetch misses\000legacy-c... */
	{ 119234 },
	/* node-prefetches-misses\000legacy cache\000Local memory prefetch misses\000legacy... */
	{ 119133 },
	/* node-prefetches-ops\000legacy cache\000Local memory prefetch accesses\000legacy-... */
	{ 118934 },
	/* node-prefetches-reference\000legacy cache\000Local memory prefetch accesses\000l... */
	{ 118830 },
	/* node-prefetches-refs\000legacy cache\000Local memory prefetch accesses\000legacy... */
	{ 118731 },
	/* node-read\000legacy cache\000Local memory read accesses\000legacy-cache-config=6... */
	{ 115422 },
	/* node-read-access\000legacy cache\000Local memory read accesses\000legacy-cache-c... */
	{ 115761 },
	/* node-read-miss\000legacy cache\000Local memory read misses\000legacy-cache-confi... */
	{ 115939 },
	/* node-read-misses\000legacy cache\000Local memory read misses\000legacy-cache-con... */
	{ 115848 },
	/* node-read-ops\000legacy cache\000Local memory read accesses\000legacy-cache-conf... */
	{ 115677 },
	/* node-read-reference\000legacy cache\000Local memory read accesses\000legacy-cach... */
	{ 115587 },
	/* node-read-refs\000legacy cache\000Local memory read accesses\000legacy-cache-con... */
	{ 115502 },
	/* node-reference\000legacy cache\000Local memory read accesses\000legacy-cache-con... */
	{ 120889 },
	/* node-refs\000legacy cache\000Local memory read accesses\000legacy-cache-config=6... */
	{ 120809 },
	/* node-speculative-load\000legacy cache\000Local memory prefetch accesses\000legac... */
	{ 120071 },
	/* node-speculative-load-access\000legacy cache\000Local memory prefetch accesses\0... */
	{ 120490 },
	/* node-speculative-load-miss\000legacy cache\000Local memory prefetch misses\000le... */
	{ 120704 },
	/* node-speculative-load-misses\000legacy cache\000Local memory prefetch misses\000... */
	{ 120597 },
	/* node-speculative-load-ops\000legacy cache\000Local memory prefetch accesses\000l... */
	{ 120386 },
	/* node-speculative-load-reference\000legacy cache\000Local memory prefetch accesse... */
	{ 120276 },
	/* node-speculative-load-refs\000legacy cache\000Local memory prefetch accesses\000... */
	{ 120171 },
	/* node-speculative-read\000legacy cache\000Local memory prefetch accesses\000legac... */
	{ 119333 },
	/* node-speculative-read-access\000legacy cache\000Local memory prefetch accesses\0... */
	{ 119752 },
	/* node-speculative-read-miss\000legacy cache\000Local memory prefetch misses\000le... */
	{ 119966 },
	/* node-speculative-read-misses\000legacy cache\000Local memory prefetch misses\000... */
	{ 119859 },
	/* node-speculative-read-ops\000legacy cache\000Local memory prefetch accesses\000l... */
	{ 119648 },
	/* node-speculative-read-reference\000legacy cache\000Local memory prefetch accesse... */
	{ 119538 },
	/* node-speculative-read-refs\000legacy cache\000Local memory prefetch accesses\000... */
	{ 119433 },
	/* node-store\000legacy cache\000Local memory write accesses\000legacy-cache-config... */
	{ 116028 },
	/* node-store-access\000legacy cache\000Local memory write accesses\000legacy-cache... */
	{ 116391 },
	/* node-store-miss\000legacy cache\000Local memory write misses\000legacy-cache-con... */
	{ 116577 },
	/* node-store-misses\000legacy cache\000Local memory write misses\000legacy-cache-c... */
	{ 116484 },
	/* node-store-ops\000legacy cache\000Local memory write accesses\000legacy-cache-co... */
	{ 116301 },
	/* node-store-reference\000legacy cache\000Local memory write accesses\000legacy-ca... */
	{ 116205 },
	/* node-store-refs\000legacy cache\000Local memory write accesses\000legacy-cache-c... */
	{ 116114 },
	/* node-stores\000legacy cache\000Local memory write accesses\000legacy-cache-confi... */
	{ 116668 },
	/* node-stores-access\000legacy cache\000Local memory write accesses\000legacy-cach... */
	{ 117035 },
	/* node-stores-miss\000legacy cache\000Local memory write misses\000legacy-cache-co... */
	{ 117223 },
	/* node-stores-misses\000legacy cache\000Local memory write misses\000legacy-cache-... */
	{ 117129 },
	/* node-stores-ops\000legacy cache\000Local memory write accesses\000legacy-cache-c... */
	{ 116944 },
	/* node-stores-reference\000legacy cache\000Local memory write accesses\000legacy-c... */
	{ 116847 },
	/* node-stores-refs\000legacy cache\000Local memory write accesses\000legacy-cache-... */
	{ 116755 },
	/* node-write\000legacy cache\000Local memory write accesses\000legacy-cache-config... */
	{ 117315 },
	/* node-write-access\000legacy cache\000Local memory write accesses\000legacy-cache... */
	{ 117678 },
	/* node-write-miss\000legacy cache\000Local memory write misses\000legacy-cache-con... */
	{ 117864 },
	/* node-write-misses\000legacy cache\000Local memory write misses\000legacy-cache-c... */
	{ 117771 },
	/* node-write-ops\000legacy cache\000Local memory write accesses\000legacy-cache-co... */
	{ 117588 },
	/* node-write-reference\000legacy cache\000Local memory write accesses\000legacy-ca... */
	{ 117492 },
	/* node-write-refs\000legacy cache\000Local memory write accesses\000legacy-cache-c... */
	{ 117401 },
	/* ref-cycles\000legacy hardware\000Total cycles; not affected by CPU frequency sca... */
	{ 123400 },
	/* stalled-cycles-backend\000legacy hardware\000Stalled cycles during retirement [T... */
	{ 123094 },
	/* stalled-cycles-frontend\000legacy hardware\000Stalled cycles during issue [This ... */
	{ 122795 },
};
static const struct compact_pmu_event pmu_events__common_software[] = {
	/* alignment-faults\000software\000Number of kernel handled memory alignment faults... */
	{ 124563 },
	/* bpf-output\000software\000An event used by BPF programs to write to the perf rin... */
	{ 124862 },
	/* cgroup-switches\000software\000Number of context switches to a task in a differe... */
	{ 124964 },
	/* context-switches\000software\000Number of context switches [This event is an ali... */
	{ 123885 },
	/* cpu-clock\000software\000Per-CPU high-resolution timer based event\000config=0\0... */
	{ 123521 },
	/* cpu-migrations\000software\000Number of times a process has migrated to a new CP... */
	{ 124087 },
	/* cs\000software\000Number of context switches [This event is an alias of context-... */
	{ 123986 },
	/* dummy\000software\000A placeholder event that doesn't count anything\000config=9... */
	{ 124782 },
	/* emulation-faults\000software\000Number of kernel handled unimplemented instructi... */
	{ 124655 },
	/* faults\000software\000Number of page faults [This event is an alias of page-faul... */
	{ 123695 },
	/* major-faults\000software\000Number of major page faults. Major faults require I/... */
	{ 124460 },
	/* migrations\000software\000Number of times a process has migrated to a new CPU [T... */
	{ 124219 },
	/* minor-faults\000software\000Number of minor page faults. Minor faults don't requ... */
	{ 124351 },
	/* page-faults\000software\000Number of page faults [This event is an alias of faul... */
	{ 123790 },
	/* task-clock\000software\000Per-task high-resolution timer based event\000config=1... */
	{ 123607 },
};
static const struct compact_pmu_event pmu_events__common_tool[] = {
	/* core_wide\000tool\0001 if not SMT, if SMT are events being gathered on all SMT t... */
	{ 126205 },
	/* duration_time\000tool\000Wall clock interval time in nanoseconds\000config=1\000... */
	{ 125072 },
	/* has_pmem\000tool\0001 if persistent memory installed otherwise 0\000config=4\000... */
	{ 125286 },
	/* num_cores\000tool\000Number of cores. A core consists of 1 or more thread, with ... */
	{ 125362 },
	/* num_cpus\000tool\000Number of logical Linux CPUs. There may be multiple such CPU... */
	{ 125507 },
	/* num_cpus_online\000tool\000Number of online logical Linux CPUs. There may be mul... */
	{ 125610 },
	/* num_dies\000tool\000Number of dies. Each die has 1 or more cores\000config=8\000... */
	{ 125727 },
	/* num_packages\000tool\000Number of packages. Each package has 1 or more die\000co... */
	{ 125803 },
	/* slots\000tool\000Number of functional units that in parallel can execute parts o... */
	{ 125889 },
	/* smt_on\000tool\0001 if simultaneous multithreading (aka hyperthreading) is enabl... */
	{ 125999 },
	/* system_time\000tool\000System/kernel time in nanoseconds\000config=3\000\00000\0... */
	{ 125218 },
	/* system_tsc_freq\000tool\000The amount a Time Stamp Counter (TSC) increases per s... */
	{ 126106 },
	/* target_cpu\000tool\0001 if CPUs being analyzed, 0 if threads/processes\000config... */
	{ 126319 },
	/* user_time\000tool\000User (non-kernel) time in nanoseconds\000config=2\000\00000... */
	{ 125148 },

};

static const struct pmu_table_entry pmu_events__common[] = {
	{
		.entries = pmu_events__common_default_core,
		.num_entries = ARRAY_SIZE(pmu_events__common_default_core),
		.pmu_name = { 0 /* default_core\000 */ },
	},
	{
		.entries = pmu_events__common_software,
		.num_entries = ARRAY_SIZE(pmu_events__common_software),
		.pmu_name = { 123512 /* software\000 */ },
	},
	{
		.entries = pmu_events__common_tool,
		.num_entries = ARRAY_SIZE(pmu_events__common_tool),
		.pmu_name = { 125067 /* tool\000 */ },
	},
};

static const struct compact_pmu_event pmu_metrics__common_default_core[] = {
	/* CPUs_utilized\000Default\000(software@cpu\\-clock\\,name\\=cpu\\-clock@ if #targ... */
	{ 127956 },
	/* backend_cycles_idle\000Default\000(stalled\\-cycles\\-backend / cpu\\-cycles if ... */
	{ 129583 },
	/* branch_frequency\000Default\000branches / (software@cpu\\-clock\\,name\\=cpu\\-c... */
	{ 129933 },
	/* branch_miss_rate\000Default\000branch\\-misses / branches\000branch_miss_rate > ... */
	{ 130113 },
	/* cs_per_second\000Default\000software@context\\-switches\\,name\\=context\\-switc... */
	{ 128142 },
	/* cycles_frequency\000Default\000cpu\\-cycles / (software@cpu\\-clock\\,name\\=cpu... */
	{ 129757 },
	/* dtlb_miss_rate\000Default3\000dTLB\\-load\\-misses / dTLB\\-loads\000dtlb_miss_r... */
	{ 130549 },
	/* frontend_cycles_idle\000Default\000(stalled\\-cycles\\-frontend / cpu\\-cycles i... */
	{ 129404 },
	/* insn_per_cycle\000Default\000instructions / cpu\\-cycles\000insn_per_cycle < 1\0... */
	{ 128866 },
	/* itlb_miss_rate\000Default3\000iTLB\\-load\\-misses / iTLB\\-loads\000itlb_miss_r... */
	{ 130655 },
	/* l1_prefetch_miss_rate\000Default4\000L1\\-dcache\\-prefetch\\-misses / L1\\-dcac... */
	{ 130761 },
	/* l1d_miss_rate\000Default2\000L1\\-dcache\\-load\\-misses / L1\\-dcache\\-loads\0... */
	{ 130217 },
	/* l1i_miss_rate\000Default3\000L1\\-icache\\-load\\-misses / L1\\-icache\\-loads\0... */
	{ 130434 },
	/* llc_miss_rate\000Default2\000LLC\\-load\\-misses / LLC\\-loads\000llc_miss_rate ... */
	{ 130333 },
	/* migrations_per_second\000Default\000software@cpu\\-migrations\\,name\\=cpu\\-mig... */
	{ 128375 },
	/* page_faults_per_second\000Default\000software@page\\-faults\\,name\\=page\\-faul... */
	{ 128635 },
	/* stalled_cycles_per_instruction\000Default\000(max(stalled\\-cycles\\-frontend, s... */
	{ 128979 },

};

static const struct pmu_table_entry pmu_metrics__common[] = {
	{
		.entries = pmu_metrics__common_default_core,
		.num_entries = ARRAY_SIZE(pmu_metrics__common_default_core),
		.pmu_name = { 0 /* default_core\000 */ },
	},
};

static const struct compact_pmu_event pmu_events__test_soc_cpu_default_core[] = {
	/* bp_l1_btb_correct\000branch\000L1 BTB Correction\000event=0x8a\000\00000\000\000... */
	{ 126403 },
	/* bp_l2_btb_correct\000branch\000L2 BTB Correction\000event=0x8b\000\00000\000\000... */
	{ 126465 },
	/* dispatch_blocked.any\000other\000Memory cluster signals to block micro-op dispat... */
	{ 126727 },
	/* eist_trans\000other\000Number of Enhanced Intel SpeedStep(R) Technology (EIST) t... */
	{ 126860 },
	/* l3_cache_rd\000cache\000L3 cache access, read\000event=0x40\000\00000\000\000\00... */
	{ 126527 },
	/* segment_reg_loads.any\000other\000Number of segment register loads\000event=6,pe... */
	{ 126625 },
};
static const struct compact_pmu_event pmu_events__test_soc_cpu_hisi_sccl_ddrc[] = {
	/* uncore_hisi_ddrc.flux_wcmd\000uncore\000DDRC write commands\000event=2\000\00000... */
	{ 126993 },
};
static const struct compact_pmu_event pmu_events__test_soc_cpu_hisi_sccl_l3c[] = {
	/* uncore_hisi_l3c.rd_hit_cpipe\000uncore\000Total read hits\000event=7\000\00000\0... */
	{ 127355 },
};
static const struct compact_pmu_event pmu_events__test_soc_cpu_uncore_cbox[] = {
	/* event-hyphen\000uncore\000UNC_CBO_HYPHEN\000event=0xe0\000\00000\000\000\000\000... */
	{ 127229 },
	/* event-two-hyph\000uncore\000UNC_CBO_TWO_HYPH\000event=0xc0\000\00000\000\000\000... */
	{ 127283 },
	/* unc_cbo_xsnp_response.miss_eviction\000uncore\000A cross-core snoop resulted fro... */
	{ 127075 },
};
static const struct compact_pmu_event pmu_events__test_soc_cpu_uncore_imc[] = {
	/* uncore_imc.cache_hits\000uncore\000Total cache hits\000event=0x34\000\00000\000\... */
	{ 127538 },
};
static const struct compact_pmu_event pmu_events__test_soc_cpu_uncore_imc_free_running[] = {
	/* uncore_imc_free_running.cache_miss\000uncore\000Total cache misses\000event=0x12... */
	{ 127447 },

};

static const struct pmu_table_entry pmu_events__test_soc_cpu[] = {
	{
		.entries = pmu_events__test_soc_cpu_default_core,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_default_core),
		.pmu_name = { 0 /* default_core\000 */ },
	},
	{
		.entries = pmu_events__test_soc_cpu_hisi_sccl_ddrc,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_hisi_sccl_ddrc),
		.pmu_name = { 126978 /* hisi_sccl,ddrc\000 */ },
	},
	{
		.entries = pmu_events__test_soc_cpu_hisi_sccl_l3c,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_hisi_sccl_l3c),
		.pmu_name = { 127341 /* hisi_sccl,l3c\000 */ },
	},
	{
		.entries = pmu_events__test_soc_cpu_uncore_cbox,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_uncore_cbox),
		.pmu_name = { 127063 /* uncore_cbox\000 */ },
	},
	{
		.entries = pmu_events__test_soc_cpu_uncore_imc,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_uncore_imc),
		.pmu_name = { 127527 /* uncore_imc\000 */ },
	},
	{
		.entries = pmu_events__test_soc_cpu_uncore_imc_free_running,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_cpu_uncore_imc_free_running),
		.pmu_name = { 127423 /* uncore_imc_free_running\000 */ },
	},
};

static const struct compact_pmu_event pmu_metrics__test_soc_cpu_default_core[] = {
	/* CPI\000\0001 / IPC\000\000\000\000\000\000\000\000000 */
	{ 130909 },
	/* DCache_L2_All\000\000DCache_L2_All_Hits + DCache_L2_All_Miss\000\000\000\000\000... */
	{ 131598 },
	/* DCache_L2_All_Hits\000\000l2_rqsts.demand_data_rd_hit + l2_rqsts.pf_hit + l2_rqs... */
	{ 131368 },
	/* DCache_L2_All_Miss\000\000max(l2_rqsts.all_demand_data_rd - l2_rqsts.demand_data... */
	{ 131463 },
	/* DCache_L2_Hits\000\000d_ratio(DCache_L2_All_Hits, DCache_L2_All)\000\000\000\000... */
	{ 131663 },
	/* DCache_L2_Misses\000\000d_ratio(DCache_L2_All_Miss, DCache_L2_All)\000\000\000\0... */
	{ 131732 },
	/* Frontend_Bound_SMT\000\000idq_uops_not_delivered.core / (4 * (cpu_clk_unhalted.t... */
	{ 130996 },
	/* IPC\000group1\000inst_retired.any / cpu_clk_unhalted.thread\000\000\000\000\000\... */
	{ 130932 },
	/* L1D_Cache_Fill_BW\000\00064 * l1d.replacement / 1e9 / duration_time\000\000\000\... */
	{ 131870 },
	/* M1\000\000ipc + M2\000\000\000\000\000\000\000\000000 */
	{ 131803 },
	/* M2\000\000ipc + M1\000\000\000\000\000\000\000\000000 */
	{ 131826 },
	/* M3\000\0001 / M3\000\000\000\000\000\000\000\000000 */
	{ 131849 },
	/* cache_miss_cycles\000group1\000dcache_miss_cpi + icache_miss_cycles\000\000\000\... */
	{ 131296 },
	/* dcache_miss_cpi\000\000l1d\\-loads\\-misses / inst_retired.any\000\000\000\000\0... */
	{ 131163 },
	/* icache_miss_cycles\000\000l1i\\-loads\\-misses / inst_retired.any\000\000\000\00... */
	{ 131228 },

};

static const struct pmu_table_entry pmu_metrics__test_soc_cpu[] = {
	{
		.entries = pmu_metrics__test_soc_cpu_default_core,
		.num_entries = ARRAY_SIZE(pmu_metrics__test_soc_cpu_default_core),
		.pmu_name = { 0 /* default_core\000 */ },
	},
};

static const struct compact_pmu_event pmu_events__test_soc_sys_uncore_sys_ccn_pmu[] = {
	/* sys_ccn_pmu.read_cycles\000uncore\000ccn read-cycles event\000config=0x2c\0000x0... */
	{ 127717 },
};
static const struct compact_pmu_event pmu_events__test_soc_sys_uncore_sys_cmn_pmu[] = {
	/* sys_cmn_pmu.hnf_cache_miss\000uncore\000Counts total cache misses in first looku... */
	{ 127813 },
};
static const struct compact_pmu_event pmu_events__test_soc_sys_uncore_sys_ddr_pmu[] = {
	/* sys_ddr_pmu.write_cycles\000uncore\000ddr write-cycles event\000event=0x2b\000v8... */
	{ 127622 },

};

static const struct pmu_table_entry pmu_events__test_soc_sys[] = {
	{
		.entries = pmu_events__test_soc_sys_uncore_sys_ccn_pmu,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_sys_uncore_sys_ccn_pmu),
		.pmu_name = { 127698 /* uncore_sys_ccn_pmu\000 */ },
	},
	{
		.entries = pmu_events__test_soc_sys_uncore_sys_cmn_pmu,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_sys_uncore_sys_cmn_pmu),
		.pmu_name = { 127794 /* uncore_sys_cmn_pmu\000 */ },
	},
	{
		.entries = pmu_events__test_soc_sys_uncore_sys_ddr_pmu,
		.num_entries = ARRAY_SIZE(pmu_events__test_soc_sys_uncore_sys_ddr_pmu),
		.pmu_name = { 127603 /* uncore_sys_ddr_pmu\000 */ },
	},
};


/* Struct used to make the PMU event table implementation opaque to callers. */
struct pmu_events_table {
	const struct pmu_table_entry *pmus;
	uint32_t num_pmus;
};

/* Struct used to make the PMU metric table implementation opaque to callers. */
struct pmu_metrics_table {
	const char *name;
	const struct pmu_table_entry *pmus;
	uint32_t num_pmus;
};

/*
 * Map a CPU to its table of PMU events. The CPU is identified by the
 * cpuid field, which is an arch-specific identifier for the CPU.
 * The identifier specified in tools/perf/pmu-events/arch/xxx/mapfile
 * must match the get_cpuid_str() in tools/perf/arch/xxx/util/header.c)
 *
 * The  cpuid can contain any character other than the comma.
 */
struct pmu_events_map {
	const char *arch;
	const char *cpuid;
	struct pmu_events_table event_table;
	struct pmu_metrics_table metric_table;
};

/*
 * Global table mapping each known CPU for the architecture to its
 * table of PMU events.
 */
static const struct pmu_events_map pmu_events_map[] = {
{
	.arch = "common",
	.cpuid = "common",
	.event_table = {
		.pmus = pmu_events__common,
		.num_pmus = ARRAY_SIZE(pmu_events__common),
	},
	.metric_table = {
		.name = "common",
		.pmus = pmu_metrics__common,
		.num_pmus = ARRAY_SIZE(pmu_metrics__common),
	},
},
{
	.arch = "testarch",
	.cpuid = "testcpu",
	.event_table = {
		.pmus = pmu_events__test_soc_cpu,
		.num_pmus = ARRAY_SIZE(pmu_events__test_soc_cpu),
	},
	.metric_table = {
		.name = "test_soc_cpu",
		.pmus = pmu_metrics__test_soc_cpu,
		.num_pmus = ARRAY_SIZE(pmu_metrics__test_soc_cpu),
	}
},
{
	.arch = 0,
	.cpuid = 0,
	.event_table = { 0, 0 },
	.metric_table = { 0 },
}
};

struct pmu_sys_events {
	const char *name;
	struct pmu_events_table event_table;
	struct pmu_metrics_table metric_table;
};

static const struct pmu_sys_events pmu_sys_event_tables[] = {
	{
		.event_table = {
			.pmus = pmu_events__test_soc_sys,
			.num_pmus = ARRAY_SIZE(pmu_events__test_soc_sys)
		},
		.name = "pmu_events__test_soc_sys",
	},
	{
		.event_table = { 0, 0 },
		.metric_table = { 0 },
	},
};

static void decompress_event(int offset, struct pmu_event *pe)
{
	const char *p = &big_c_string[offset];

	pe->name = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->topic = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->desc = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->event = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->compat = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->deprecated = *p - '0';
	p++;
	pe->perpkg = *p - '0';
	p++;
	pe->unit = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->retirement_latency_mean = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->retirement_latency_min = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->retirement_latency_max = (*p == '\0' ? NULL : p);
	while (*p++);
	pe->long_desc = (*p == '\0' ? NULL : p);
}

static void decompress_metric(int offset, struct pmu_metric *pm)
{
	const char *p = &big_c_string[offset];

	pm->metric_name = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->metric_group = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->metric_expr = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->metric_threshold = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->desc = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->long_desc = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->unit = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->compat = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->metricgroup_no_group = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->default_metricgroup_name = (*p == '\0' ? NULL : p);
	while (*p++);
	pm->aggr_mode = *p - '0';
	p++;
	pm->event_grouping = *p - '0';
	p++;
	pm->default_show_events = *p - '0';
}

static int pmu_events_table__for_each_event_pmu(const struct pmu_events_table *table,
						const struct pmu_table_entry *pmu,
						pmu_event_iter_fn fn,
						void *data)
{
	int ret;
	struct pmu_event pe = {
		.pmu = &big_c_string[pmu->pmu_name.offset],
	};

	for (uint32_t i = 0; i < pmu->num_entries; i++) {
		decompress_event(pmu->entries[i].offset, &pe);
		if (!pe.name)
			continue;
		ret = fn(&pe, table, data);
		if (ret)
			return ret;
	}
	return 0;
 }

static int pmu_events_table__find_event_pmu(const struct pmu_events_table *table,
					    const struct pmu_table_entry *pmu,
					    const char *name,
					    pmu_event_iter_fn fn,
					    void *data)
{
	struct pmu_event pe = {
		.pmu = &big_c_string[pmu->pmu_name.offset],
	};
	int low = 0, high = pmu->num_entries - 1;

	while (low <= high) {
		int cmp, mid = (low + high) / 2;

		decompress_event(pmu->entries[mid].offset, &pe);

		if (!pe.name && !name)
			goto do_call;

		if (!pe.name && name) {
			low = mid + 1;
			continue;
		}
		if (pe.name && !name) {
			high = mid - 1;
			continue;
		}

		cmp = strcasecmp(pe.name, name);
		if (cmp < 0) {
			low = mid + 1;
			continue;
		}
		if (cmp > 0) {
			high = mid - 1;
			continue;
		}
  do_call:
		return fn ? fn(&pe, table, data) : 0;
	}
	return PMU_EVENTS__NOT_FOUND;
}

int pmu_events_table__for_each_event(const struct pmu_events_table *table,
				    struct perf_pmu *pmu,
				    pmu_event_iter_fn fn,
				    void *data)
{
	if (!table)
		return 0;
	for (size_t i = 0; i < table->num_pmus; i++) {
		const struct pmu_table_entry *table_pmu = &table->pmus[i];
		const char *pmu_name = &big_c_string[table_pmu->pmu_name.offset];
		int ret;

		if (pmu && !perf_pmu__name_wildcard_match(pmu, pmu_name))
			continue;

		ret = pmu_events_table__for_each_event_pmu(table, table_pmu, fn, data);
		if (ret)
			return ret;
	}
	return 0;
}

int pmu_events_table__find_event(const struct pmu_events_table *table,
				 struct perf_pmu *pmu,
				 const char *name,
				 pmu_event_iter_fn fn,
				 void *data)
{
	if (!table)
		return PMU_EVENTS__NOT_FOUND;
	for (size_t i = 0; i < table->num_pmus; i++) {
		const struct pmu_table_entry *table_pmu = &table->pmus[i];
		const char *pmu_name = &big_c_string[table_pmu->pmu_name.offset];
		int ret;

		if (pmu && !perf_pmu__name_wildcard_match(pmu, pmu_name))
			continue;

		ret = pmu_events_table__find_event_pmu(table, table_pmu, name, fn, data);
		if (ret != PMU_EVENTS__NOT_FOUND)
			return ret;
	}
	return PMU_EVENTS__NOT_FOUND;
}

size_t pmu_events_table__num_events(const struct pmu_events_table *table, struct perf_pmu *pmu)
{
	size_t count = 0;

	if (!table)
		return 0;
	for (size_t i = 0; i < table->num_pmus; i++) {
		const struct pmu_table_entry *table_pmu = &table->pmus[i];
		const char *pmu_name = &big_c_string[table_pmu->pmu_name.offset];

		if (perf_pmu__name_wildcard_match(pmu, pmu_name))
			count += table_pmu->num_entries;
	}
	return count;
}

static int pmu_metrics_table__for_each_metric_pmu(const struct pmu_metrics_table *table,
						const struct pmu_table_entry *pmu,
						pmu_metric_iter_fn fn,
						void *data)
{
	int ret;
	struct pmu_metric pm = {
		.pmu = &big_c_string[pmu->pmu_name.offset],
	};

	for (uint32_t i = 0; i < pmu->num_entries; i++) {
		decompress_metric(pmu->entries[i].offset, &pm);
		if (!pm.metric_expr)
			continue;
		ret = fn(&pm, table, data);
		if (ret)
			return ret;
	}
	return 0;
}

static int pmu_metrics_table__find_metric_pmu(const struct pmu_metrics_table *table,
					    const struct pmu_table_entry *pmu,
					    const char *metric,
					    pmu_metric_iter_fn fn,
					    void *data)
{
	struct pmu_metric pm = {
		.pmu = &big_c_string[pmu->pmu_name.offset],
	};
	int low = 0, high = pmu->num_entries - 1;

	while (low <= high) {
		int cmp, mid = (low + high) / 2;

		decompress_metric(pmu->entries[mid].offset, &pm);

		if (!pm.metric_name && !metric)
			goto do_call;

		if (!pm.metric_name && metric) {
			low = mid + 1;
			continue;
		}
		if (pm.metric_name && !metric) {
			high = mid - 1;
			continue;
		}

		cmp = strcmp(pm.metric_name, metric);
		if (cmp < 0) {
			low = mid + 1;
			continue;
		}
		if (cmp > 0) {
			high = mid - 1;
			continue;
		}
  do_call:
		return fn ? fn(&pm, table, data) : 0;
	}
	return PMU_METRICS__NOT_FOUND;
}

int pmu_metrics_table__for_each_metric(const struct pmu_metrics_table *table,
				     pmu_metric_iter_fn fn,
				     void *data)
{
	if (!table)
		return 0;
	for (size_t i = 0; i < table->num_pmus; i++) {
		int ret = pmu_metrics_table__for_each_metric_pmu(table, &table->pmus[i], fn, data);

		if (ret)
			return ret;
	}
	return 0;
}

int pmu_metrics_table__find_metric(const struct pmu_metrics_table *table,
				 struct perf_pmu *pmu,
				 const char *metric,
				 pmu_metric_iter_fn fn,
				 void *data)
{
	if (!table)
		return 0;
	for (size_t i = 0; i < table->num_pmus; i++) {
		const struct pmu_table_entry *table_pmu = &table->pmus[i];
		const char *pmu_name = &big_c_string[table_pmu->pmu_name.offset];
		int ret;

		if (pmu && !perf_pmu__name_wildcard_match(pmu, pmu_name))
			continue;

		ret = pmu_metrics_table__find_metric_pmu(table, table_pmu, metric, fn, data);
		if (ret != PMU_METRICS__NOT_FOUND)
			return ret;
	}
	return PMU_METRICS__NOT_FOUND;
}

static const struct pmu_events_map *map_for_cpu(struct perf_cpu cpu)
{
	static struct {
		const struct pmu_events_map *map;
		struct perf_cpu cpu;
	} last_result;
	static struct {
		const struct pmu_events_map *map;
		char *cpuid;
	} last_map_search;
	static bool has_last_result, has_last_map_search;
	const struct pmu_events_map *map = NULL;
	char *cpuid = NULL;
	size_t i;

	if (has_last_result && last_result.cpu.cpu == cpu.cpu)
		return last_result.map;

	cpuid = get_cpuid_allow_env_override(cpu);

	/*
	 * On some platforms which uses cpus map, cpuid can be NULL for
	 * PMUs other than CORE PMUs.
	 */
	if (!cpuid)
		goto out_update_last_result;

	if (has_last_map_search && !strcmp(last_map_search.cpuid, cpuid)) {
		map = last_map_search.map;
		free(cpuid);
	} else {
		i = 0;
		for (;;) {
			map = &pmu_events_map[i++];

			if (!map->arch) {
				map = NULL;
				break;
			}

			if (!strcmp_cpuid_str(map->cpuid, cpuid))
				break;
		}
		free(last_map_search.cpuid);
		last_map_search.cpuid = cpuid;
		last_map_search.map = map;
		has_last_map_search = true;
	}
out_update_last_result:
	last_result.cpu = cpu;
	last_result.map = map;
	has_last_result = true;
	return map;
}

static const struct pmu_events_map *map_for_pmu(struct perf_pmu *pmu)
{
	struct perf_cpu cpu = { -1 };

	if (pmu) {
		for (size_t i = 0; i < ARRAY_SIZE(pmu_events__common); i++) {
			const char *pmu_name = &big_c_string[pmu_events__common[i].pmu_name.offset];

			if (!strcmp(pmu_name, pmu->name)) {
				const struct pmu_events_map *map = &pmu_events_map[0];

				while (strcmp("common", map->arch))
					map++;
				return map;
			}
		}
		cpu = perf_cpu_map__min(pmu->cpus);
	}
	return map_for_cpu(cpu);
}

const struct pmu_events_table *perf_pmu__find_events_table(struct perf_pmu *pmu)
{
	const struct pmu_events_map *map = map_for_pmu(pmu);

	if (!map)
		return NULL;

	if (!pmu)
		return &map->event_table;

	for (size_t i = 0; i < map->event_table.num_pmus; i++) {
		const struct pmu_table_entry *table_pmu = &map->event_table.pmus[i];
		const char *pmu_name = &big_c_string[table_pmu->pmu_name.offset];

		if (perf_pmu__name_wildcard_match(pmu, pmu_name))
			return &map->event_table;
	}
	return NULL;
}

const struct pmu_events_table *perf_pmu__default_core_events_table(void)
{
	int i = 0;

	for (;;) {
		const struct pmu_events_map *map = &pmu_events_map[i++];

		if (!map->arch)
			break;

		if (!strcmp(map->cpuid, "common"))
			return &map->event_table;
	}
	return NULL;
}

const struct pmu_metrics_table *pmu_metrics_table__find(void)
{
	struct perf_cpu cpu = { -1 };
	const struct pmu_events_map *map = map_for_cpu(cpu);

	return map ? &map->metric_table : NULL;
}

const struct pmu_metrics_table *pmu_metrics_table__default(void)
{
	int i = 0;

	for (;;) {
		const struct pmu_events_map *map = &pmu_events_map[i++];

		if (!map->arch)
			break;

		if (!strcmp(map->cpuid, "common"))
			return &map->metric_table;
	}
	return NULL;
}

const struct pmu_events_table *find_core_events_table(const char *arch, const char *cpuid)
{
	for (const struct pmu_events_map *tables = &pmu_events_map[0];
	     tables->arch;
	     tables++) {
		if (!strcmp(tables->arch, arch) && !strcmp_cpuid_str(tables->cpuid, cpuid))
			return &tables->event_table;
	}
	return NULL;
}

const struct pmu_metrics_table *find_core_metrics_table(const char *arch, const char *cpuid)
{
	for (const struct pmu_events_map *tables = &pmu_events_map[0];
	     tables->arch;
	     tables++) {
		if (!strcmp(tables->arch, arch) && !strcmp_cpuid_str(tables->cpuid, cpuid))
			return &tables->metric_table;
	}
	return NULL;
}

int pmu_for_each_core_event(pmu_event_iter_fn fn, void *data)
{
	for (const struct pmu_events_map *tables = &pmu_events_map[0];
	     tables->arch;
	     tables++) {
		int ret = pmu_events_table__for_each_event(&tables->event_table,
							   /*pmu=*/NULL, fn, data);

		if (ret)
			return ret;
	}
	return 0;
}

int pmu_for_each_core_metric(pmu_metric_iter_fn fn, void *data)
{
	for (const struct pmu_events_map *tables = &pmu_events_map[0];
	     tables->arch;
	     tables++) {
		int ret = pmu_metrics_table__for_each_metric(&tables->metric_table, fn, data);

		if (ret)
			return ret;
	}
	return 0;
}

const struct pmu_events_table *find_sys_events_table(const char *name)
{
	for (const struct pmu_sys_events *tables = &pmu_sys_event_tables[0];
	     tables->name;
	     tables++) {
		if (!strcmp(tables->name, name))
			return &tables->event_table;
	}
	return NULL;
}

int pmu_for_each_sys_event(pmu_event_iter_fn fn, void *data)
{
	for (const struct pmu_sys_events *tables = &pmu_sys_event_tables[0];
	     tables->name;
	     tables++) {
		int ret = pmu_events_table__for_each_event(&tables->event_table,
							   /*pmu=*/NULL, fn, data);

		if (ret)
			return ret;
	}
	return 0;
}

int pmu_for_each_sys_metric(pmu_metric_iter_fn fn, void *data)
{
	for (const struct pmu_sys_events *tables = &pmu_sys_event_tables[0];
	     tables->name;
	     tables++) {
		int ret = pmu_metrics_table__for_each_metric(&tables->metric_table, fn, data);

		if (ret)
			return ret;
	}
	return 0;
}
/* clang-format on */

const char *pmu_metrics_table__name(const struct pmu_metrics_table *table)
{
	return table ? table->name : NULL;
}

int pmu_metrics_table__iterate_tables(pmu_metrics_table_iter_t fn, void *data)
{
	size_t i;
	int ret;

	for (i = 0; pmu_events_map[i].cpuid; i++) {
		size_t j;
		bool found = false;

		if (!pmu_events_map[i].metric_table.pmus)
			continue;
		for (j = 0; j < i; j++) {
			if (pmu_events_map[j].metric_table.pmus ==
			    pmu_events_map[i].metric_table.pmus) {
				found = true;
				break;
			}
		}
		if (found)
			continue;
		ret = fn(&pmu_events_map[i].metric_table, data);
		if (ret)
			return ret;
	}
	for (i = 0; pmu_sys_event_tables[i].name; i++) {
		if (!pmu_sys_event_tables[i].metric_table.pmus)
			continue;
		ret = fn(&pmu_sys_event_tables[i].metric_table, data);
		if (ret)
			return ret;
	}
	return 0;
}

static const int metricgroups[][2] = {

};

const char *describe_metricgroup(const char *group)
{
	int low = 0, high = (int)ARRAY_SIZE(metricgroups) - 1;

	while (low <= high) {
		int mid = (low + high) / 2;
		const char *mgroup = &big_c_string[metricgroups[mid][0]];
		int cmp = strcmp(mgroup, group);

		if (cmp == 0)
			return &big_c_string[metricgroups[mid][1]];
		else if (cmp < 0)
			low = mid + 1;
		else
			high = mid - 1;
	}
	return NULL;
}
